// wzAIEngine.cpp - AI engine, alpha-beta search
// Copyright (C) 2006  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzAIEngine.h"
#include "wzGame.h"
#include "wzAIBoard.h"
#include "wzReferee.h"
#include "wzEvaluator.h"
#include <algorithm>


// initialize static data members
wzEvaluator wzAIEngine::m_Evaluator;


////////////////////////////
wzAIEngine::wzAIEngine( short Level )
  : m_pGame( NULL ), m_pBoard( NULL ) 
{
    if( Level == 1 )
        m_Depth = 2;
    else if( Level == 2 )
        m_Depth = 4;
}

/////////////////////
wzAIEngine::~wzAIEngine()
{ 
    if( m_pBoard != NULL )
    {
        delete m_pBoard;
        m_pBoard = NULL;
    }
}

/////////////////////////////////////////////////////////////
void wzAIEngine::BestMove( wzGame* pGame, wzBoardBase* pBoard, wzMove& mBest )
{
    if( pBoard->IsEmpty() )
    {
        mBest.X = 7;
        mBest.Y = 7;
        return;
    }

    if( m_pBoard != NULL )
    {
        delete m_pBoard;
        m_pBoard = NULL;
    }

    m_pGame   = pGame;
    m_MyColor = mBest.Stone;

    m_pBoard = new wzAIBoard( *pBoard );  // make a copy

    wzAIMove NullMove;
    AlphaBeta( 0, -wzInfinite, wzInfinite, m_MyColor, true, NullMove, mBest );
}

//////////////////
// MaxMini with alpha-beta cut-offs, sorted child nodes
////////////////////////////////////////////////////////////////////////////
long long wzAIEngine::AlphaBeta( short depth, long long alpha, long long beta, 
                                 wzSTONE stone, bool bMax, wzAIMove& amLast, 
                                 wzMove& mBest )
{
    if( !m_pGame->KeepRunning() )  return 0LL;

#if 1
    if( bMax )
        if( alpha >= 1000000000000LL ) return alpha;
    else
        if( beta <= -1000000000000LL ) return beta;
#endif

    // leaf node?
    if( depth > 0 && IsLeaf( depth, amLast ) )
    {
        // full evaluation
        return m_Evaluator.Evaluate( *m_pBoard, m_MyColor );
        // evaluation based on last move
        //return amLast.Score; 
    }

    // alpha and beta are local to each node
    long long  alpha0 = alpha;
    long long  beta0  = beta;
    long long  score0 = 0LL;
    short      depth0 = ++depth;

    // generate all child nodes
    wzAIMoves moves = GenMoves( stone );

    // sort 
    if( bMax )
        sort( moves.begin(), moves.end(), wzDescendMoves() );
    else
        sort( moves.begin(), moves.end(), wzAscendMoves() );

    wzAIMove am;
    wzMove mNull;
    for( wzAIMovesIterator it = moves.begin(); it != moves.end(); ++it ) 
    {
        if( !m_pGame->KeepRunning() )  return 0LL;

        am = *it;
        m_pBoard->Set( am.Move ); 
        score0 = AlphaBeta( depth0, alpha0, beta0, Toggle( stone ), 
                            !bMax, am, mNull );
        // undo move
        m_pBoard->Reset( am.Move );

        if( bMax )
        {
            if( alpha0 < score0 )
            {
                alpha0 = score0;
                mBest.X = am.Move.X;
                mBest.Y = am.Move.Y;
            }

            // cut-off
            if( alpha0 >= beta0 )  return alpha0;
        }
        else
        {
            if(  beta0 > score0 )
            {
                beta0 = score0;
                mBest.X = am.Move.X;
                mBest.Y = am.Move.Y;
            }

            // cut-off
            if( alpha0 >= beta0 )  return beta0;
        }
    }

    if( bMax )
        return alpha0;
    else
        return beta0;
}

////////////////////////
bool wzAIEngine::IsLeaf( short depth, wzAIMove& am )
{
    if( depth >= m_Depth )
        return true;
    else if( wzGame::m_Referee.Winner( *m_pBoard, am.Move ) != wzNONE )
        return true;
    else if( m_pBoard->IsFull() )
        return true;
    else
        return false;
}


////////////////////////////////////////
// generate and sort all possible moves
////////////////////////////////////////////////////////////////
wzAIEngine::wzAIMoves wzAIEngine::GenMoves( wzSTONE stone )
{
    const wzBitBoard& fields = m_pBoard->GetFields();
    wzAIMoves moves;
    for( short y = 0; y < wzDimension; y++ )
    {
        for( short x = 0; x < wzDimension; x++ )
        {
            if( !m_pGame->KeepRunning() )
                return moves;

            if( !fields.IsEmpty( x, y ) && m_pBoard->Set( x, y, stone ) )
            {
                wzAIMove am;
                am.Move.X = x;
                am.Move.Y = y;
                am.Move.Stone = stone;
#if 1
                // evaluation based on last move
                am.Score = m_Evaluator.Evaluate( *m_pBoard, m_MyColor, x, y );
#else
                // full evaluation
                am.Score = m_Evaluator.Evaluate( *m_pBoard, m_MyColor );
#endif
                moves.push_back( am );

                m_pBoard->Reset( am.Move );  // undo moving
            }
        }
    }

    return moves;
}
