// wzBitBoard.cpp - Board represented by bits
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzBitBoard.h"


// initialize static data members
wzBitBoard::wzBitCounter wzBitBoard::m_BitCounter;

// bytes have one bit set
const wzUInt8 wzBitBoard::m_OneBit[8] = 
{
    0x80,  // 10000000
    0x40,  // 01000000 
    0x20,  // 00100000
    0x10,  // 00010000
    0x08,  // 00001000
    0x04,  // 00000100
    0x02,  // 00000010
    0x01   // 00000001
};

// bytes have 7 bits set
const wzUInt8 wzBitBoard::m_7Bits[8] = 
{
    0x7f,  // 01111111
    0xbf,  // 10111111
    0xdf,  // 11011111
    0xef,  // 11101111
    0xf7,  // 11110111 
    0xfb,  // 11111011 
    0xfd,  // 11111101
    0xfe   // 11111110
};


/////////////////////////////////////////////
wzBitBoard::wzBitBoard( const wzBitBoard& other )
{
    for( short y = 0; y < wzDimension; y++ )
    {
        m_Bits[0][y] = other.m_Bits[0][y];
        m_Bits[1][y] = other.m_Bits[1][y];
    }
}

///////////////////////////////////////////////
void wzBitBoard::Clean()
{
    for( short y = 0; y < wzDimension; y++ )
    {
        m_Bits[0][y] = 0x00;
        m_Bits[1][y] = 0x00;
    }
}

////////////////////////////////////////////////////////
wzBitBoard& wzBitBoard::operator=( const wzBitBoard& other )
{
    if( this == &other )
        return( *this );

    for( short y = 0; y < wzDimension; y++ )
    {
        m_Bits[0][y] = other.m_Bits[0][y];
        m_Bits[1][y] = other.m_Bits[1][y];
    }

    return( *this );
}

//////////////////////////////////////////////////////////
bool wzBitBoard::operator==( const wzBitBoard& other ) const
{
    if( this == &other )
        return true;

    for( short y = 0; y < wzDimension; y++ )
    {
        if( m_Bits[0][y] != other.m_Bits[0][y] )
            return false;

        if( m_Bits[1][y] != other.m_Bits[1][y] )
            return false;
    }

    return true;
}

///////////////////////////////////////////////////////////
bool wzBitBoard::operator!=( const wzBitBoard& other ) const
{
    if( this == &other )
        return false;

    for( short y = 0; y < wzDimension; y++ )
    {
        if( m_Bits[0][y] != other.m_Bits[0][y] )
            return true;

        if( m_Bits[1][y] != other.m_Bits[1][y] )
            return true;
    }

    return false;
}

////////////////////////////
// AND operator
//   Perform bitwise AND operation of two bit board, count 
//   the number of bits.
//
// This operator is the mostly called function and comsumes 
// the most CPU time (~87%).
//////////////////////////////////////////////////////////
short wzBitBoard::operator&( const wzBitBoard& other ) const
{
    short    nBits = 0;
    wzUInt8  Byte;
    for( short y = 0; y < wzDimension; y++ )
    {
        for( short iByte = 0; iByte < 2; iByte++ )
        {
            Byte = m_Bits[iByte][y] & other.m_Bits[iByte][y];
            nBits += m_BitCounter( Byte );
        }
    }

    return nBits;
}

////////////////////////////
// Addition operator
//   Perform bitwise OR operation with passed in bit board.
//////////////////////////////////////////////////////// 
void wzBitBoard::operator+=( const wzBitBoard& other )
{
    for( short y = 0; y < wzDimension; y++ )
    {
        for( short iByte = 0; iByte < 2; iByte++ )
        {
            m_Bits[iByte][y] |= other.m_Bits[iByte][y];
        }
    }
}

////////////////////////////
bool wzBitBoard::Set( const short& x, const short& y )
{
    short  iByte = ByteIndex( x );
    short  iBit  = BitIndex( x );

    // already set
    if( (m_Bits[iByte][y] & m_OneBit[iBit]) != 0x00 )
        return false;

    m_Bits[iByte][y] |= m_OneBit[iBit];

    return true;
}

////////////////////////////
void wzBitBoard::Reset( const short& x, const short& y )
{
    short  iByte = ByteIndex( x );
    short  iBit  = BitIndex( x );

    m_Bits[iByte][y] &= m_7Bits[iBit];
}

/////////////////////////////////////////////////////////////
bool wzBitBoard::IsEmpty( const short& x, const short& y ) const
{
    short  iByte = ByteIndex( x );
    short  iBit  = BitIndex( x );

    return( (m_Bits[iByte][y] & m_OneBit[iBit]) == 0x00 );
}

/////////////////////////////////////////////////////
wzBitBoard::wzBitCounter::wzBitCounter()
{
    // initialize the array: its index is value of the byte
    // and its value is the number of bits of the byte.
    wzUInt8  byte = 0x00;
    for( short i = 0; i < 256; i++ )
    {
        m_Count[byte] = Count( byte );
        byte += 0x01;
    }
}

////////////////////////////
// count the number of bits
/////////////////////////////////////////////////////
short wzBitBoard::wzBitCounter::Count( const wzUInt8& Byte )
{
    short nBits = 0;
    for( short i = 0; i < 8; i++ )
    {
        if( (Byte & wzBitBoard::m_OneBit[i]) != 0x00 )
            nBits++;
    }

    return nBits;
}
