// wzBoardBase.cpp - Game board base. 
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzBoardBase.h"
#include "wzBoardObserver.h"
#include "wzMove.h"

#include <algorithm>


//////////////////////////////////////
wzBoardBase::wzBoardBase( const wzBoardBase& other )
  : m_AllStones  ( other.m_AllStones ),
    m_BlackStones( other.m_BlackStones ),
    m_WhiteStones( other.m_WhiteStones ),
    m_Plies( other.m_Plies )
{
    // won't carry over observers
    m_Observers.clear();
}

/////////////////////////////////////////////////////
wzBoardBase& wzBoardBase::operator=( const wzBoardBase& other )
{
    if( this == &other )
        return( *this );

    m_AllStones   = other.m_AllStones;
    m_BlackStones = other.m_BlackStones;
    m_WhiteStones = other.m_WhiteStones;

    m_Plies = other.m_Plies;

    // won't carry over observers
    m_Observers.clear();

    return( *this );
}

//////////////////////////////////////
bool wzBoardBase::Set( const wzMove& move  )
{
    if( !m_AllStones.Set( move.X, move.Y ) )
        return false;

    if( move.Stone == wzBLACK )
        m_BlackStones.Set( move.X, move.Y );
    else
        m_WhiteStones.Set( move.X, move.Y );

    m_Plies++;

    // notify observers
#if 0
    NotifySet( move );
#else
    // save a function call
    for( wzObserversIterator it = m_Observers.begin(); 
         it != m_Observers.end(); it++ )
    {
        (*it)->Set( move );
    }
#endif

    return true;
}

//////////////////////////////////////
bool wzBoardBase::Set( short x, short y, wzSTONE stone )
{
    wzMove move;
    move.X = x; 
    move.Y = y; 
    move.Stone = stone;

    return Set( move );
}

//////////////////////////////////////
bool wzBoardBase::Reset( const wzMove& move )
{
    if( m_AllStones.IsEmpty( move.X, move.Y ) )
        return false;

    m_AllStones.Reset( move.X, move.Y );
    m_BlackStones.Reset( move.X, move.Y );
    m_WhiteStones.Reset( move.X, move.Y );

    m_Plies--;

    // notify observers
#if 0
    NotifyReset( move );
#else 
    // save a function call
    for( wzObserversIterator it = m_Observers.begin(); 
         it != m_Observers.end(); it++ )
    {
        (*it)->Reset( move );
    }
#endif

    return true;
}

//////////////////////////////////////
bool wzBoardBase::Reset( short x, short y )
{
    wzMove move;
    move.X = x; 
    move.Y = y;

    return Reset( move );
}

//////////////////////////////////////
void wzBoardBase::Clean()
{
    m_AllStones.Clean();
    m_BlackStones.Clean();
    m_WhiteStones.Clean();

    m_Plies = 0;

    // notify observers
    NotifyClean();
}

//////////////////////////////////////
wzSTONE wzBoardBase::GetStone( short x, short y ) const
{
    if( m_AllStones.IsEmpty( x, y ) )
        return wzNONE;

    if( m_BlackStones.IsEmpty( x, y ) )
        return wzWHITE;
    else
        return wzBLACK;
}

//////////////////////////////////////
void wzBoardBase::Register( wzBoardObserver* pObserver )
{
    wzObserversIterator pos = find( m_Observers.begin(), m_Observers.end(), pObserver ); 
    if( pos != m_Observers.end() )
        return;

    pObserver->Init( this );
    m_Observers.push_back( pObserver );
}

//////////////////////////////////////
void wzBoardBase::Remove( wzBoardObserver* pObserver )
{
    wzObserversIterator pos = find( m_Observers.begin(), m_Observers.end(), pObserver ); 
    if( pos != m_Observers.end() )
        m_Observers.erase( pos );
}

/////////////////////////////////////////////////////
void wzBoardBase::NotifySet( const wzMove& m )
{
    for( wzObserversIterator it = m_Observers.begin(); 
         it != m_Observers.end(); it++ )
    {
        (*it)->Set( m );
    }
}

/////////////////////////////////////////////////////
void wzBoardBase::NotifyReset( const wzMove& m  )
{
    for( wzObserversIterator it = m_Observers.begin(); 
         it != m_Observers.end(); it++ )
    {
        (*it)->Reset( m );
    }
}

/////////////////////////////////////////////////////
void wzBoardBase::NotifyClean()
{
    for( wzObserversIterator it = m_Observers.begin(); 
         it != m_Observers.end(); it++ )
    {
        (*it)->Clean();
    }
}
