// wzBoardBase.h - Game board base. 
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef __WZ_BOARD_BASE_H__
#define __WZ_BOARD_BASE_H__

#include "wzBitBoard.h"
#include <vector>

class wzBoardObserver;
struct wzMove;

////////////////////
// Base class of board, provides basic needs and is the subject
// of various observers 
////////////////////////////////////////////////////////////////////
class wzBoardBase
{
    friend class  wzEvaluator;
    friend class  wzReferee;

public:
    wzBoardBase() : m_Plies( 0 ) {}
    wzBoardBase( const wzBoardBase& other );

    virtual ~wzBoardBase() {}

    // operators
    wzBoardBase& operator=( const wzBoardBase& other );

    // inquiries
    bool     IsEmpty() const;
    bool     IsFull()  const;
    short    Plies()   const;
    bool     IsEmpty ( short x, short y ) const;
    wzSTONE  GetStone( short x, short y ) const;

    // operations
    bool  Set  ( short x, short y, wzSTONE stone );
    bool  Set  ( const wzMove& m );
    bool  Reset( short x, short y );
    bool  Reset( const wzMove& m  );
    void  Clean();

    // observer related
    void  Register( wzBoardObserver* pObserver );
    void  Remove  ( wzBoardObserver* pObserver );
    void  NotifySet( const wzMove& m );
    void  NotifyReset( const wzMove& m  );  
    void  NotifyClean();  

    static bool  Valid( const short& xy );
    static bool  Valid( const short& x, const short& y );

protected:
    // bit boards
    wzBitBoard  m_AllStones;
    wzBitBoard  m_BlackStones;
    wzBitBoard  m_WhiteStones;

    // number of moves
    short  m_Plies;

    // observer list
    typedef std::vector<wzBoardObserver*>  wzObservers;
    typedef wzObservers::iterator          wzObserversIterator;
    wzObservers  m_Observers;
};


inline bool wzBoardBase::IsEmpty() const
{
    return( m_Plies == 0 );
}

inline bool wzBoardBase::IsFull() const
{
    return( m_Plies == wzLattices );
}

inline short wzBoardBase::Plies() const
{
    return m_Plies;
}

inline bool wzBoardBase::IsEmpty( short x, short y ) const
{
    return m_AllStones.IsEmpty( x, y );
}

///////////////////////////////
// Verify x- or y-coordinate.
// 
// Functions
//     IsEmpty( x, y ), GetStone( x, y ), and
//     Set( x, y stone ), Reset( x, y )
// do not verify the x- and y-coordinate, which is intended
// to reduce overhead to the search procedure of the AI. 
// However, the verification is needed during interacting 
// with players other than the AI. 
////////////////////////////////////////////////////////////
inline bool wzBoardBase::Valid( const short& xy )
{
    return( xy >= 0 && xy < wzDimension );
}

inline bool wzBoardBase::Valid( const short& x, const short& y )
{
    return( Valid( x ) && Valid( y ) );
}

#endif
