// wzClock.cpp - Clock to count time spent by players
// Copyright (C) 2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzClock.h"

#include <sstream>
#include <iomanip>
#include <fstream>
using namespace std;


// in centi-seconds
const long wzClock::m_Limit = (MAX_HOURS*3600 + MAX_MINUTES*60 + MAX_SECONDS)*100
                              + MAX_DECISECS*10; 

////////////////////////////
void wzClock::Reset()
{
    m_Running = false;
    m_Elapsed = 0;
}

/////////////////////////////////////////
bool wzClock::Start()
{
    if( IsRunning() )
        return false;

    if( ExceedLimit() )
        return false;

    if( gettimeofday( &m_Start, NULL ) != 0 )
        return false;

    m_Running = true;
    return true;
}

////////////////////////////////////////
bool wzClock::Stop()
{
    if( !IsRunning() )
        return false;

    m_Running = false;

    if( gettimeofday( &m_Stop, NULL ) != 0 )
        return false;

    Update();
    return true;
}

////////////////////////////////////////////
void wzClock::Update()
{
    m_Elapsed += (m_Stop.tv_sec - m_Start.tv_sec)*100;
    m_Elapsed += (m_Stop.tv_usec - m_Start.tv_usec)/10000;

    if( m_Elapsed > m_Limit )
        m_Elapsed = m_Limit;
}

//////////////////////////////////////////////////////////
string wzClock::GetTime()
{
    // cache it to avoid repeating the same calculation
    static string  stime( "00:00:00.0" );

    bool bStopped = Stop();

    int  hours, minutes, seconds, decisecs;
    if( Convert( hours, minutes, seconds, decisecs ) )
    {
        stringstream  strs;
        strs << setw(2) << setfill('0') << hours   << ":" 
             << setw(2) << setfill('0') << minutes << ":" 
             << setw(2) << setfill('0') << seconds << "." 
             << setw(1) << decisecs;

        stime = strs.str();
    }

    if( bStopped ) 
        Start();

    return stime;
}

////////////////////////////////////////////////////////////////////////////
bool wzClock::Convert( int& hours, int& minutes, int& seconds, int& decisecs )
{
    // cache them to avoid repeating the same calculation
    static int  Elapsed  = 0;
    static int  Hours    = 0;
    static int  Minutes  = 0;
    static int  Seconds  = 0;
    static int  Decisecs = 0;

    int elapsed = Centi2Deci( m_Elapsed );

    // no change
    if( elapsed == Elapsed )
    {
        hours    = Hours;
        minutes  = Minutes;
        seconds  = Seconds;
        decisecs = Decisecs;

        return false;
    }    
    else;
    {
        Elapsed  = elapsed;
        hours    = Hours    =  elapsed/36000;
        minutes  = Minutes  = (elapsed - hours*36000)/600;
        seconds  = Seconds  = (elapsed - hours*36000 - minutes*600)/10;
        decisecs = Decisecs = (elapsed - hours*36000 - minutes*600 - seconds*10);

        return true;
    }    
}

//////////////////////////////////////////////////////
ifstream& operator>>( ifstream& iFile, wzClock& clock )
{
    wzUInt8  hours, minutes, seconds, decisecs;
    hours    = static_cast<wzUInt8>( iFile.get() );
    minutes  = static_cast<wzUInt8>( iFile.get() );
    seconds  = static_cast<wzUInt8>( iFile.get() );
    decisecs = static_cast<wzUInt8>( iFile.get() );

    clock.Reset();
    clock.m_Elapsed = (hours*3600 + minutes*60 + seconds)*100 + decisecs*10;

    return iFile;
}

//////////////////////////////////////////////////////
ofstream& operator<<( ofstream& oFile, wzClock& clock )
{
    bool bStopped = clock.Stop();

    int  hours, minutes, seconds, decisecs;
    clock.Convert( hours, minutes, seconds, decisecs );

    oFile << static_cast<wzUInt8>( hours );
    oFile << static_cast<wzUInt8>( minutes );
    oFile << static_cast<wzUInt8>( seconds );
    oFile << static_cast<wzUInt8>( decisecs );

    if( bStopped ) 
        clock.Start();

    return oFile;
}
