// wzGame.cpp - The game class
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzGame.h"
#include "wzBoard.h"
#include "wzGameState.h"
#include "wzGameRunner.h"
#include "wzGameReplayer.h"
#include "wzGameFile.h"
#include "wzPlayerFactory.h"
#include "wzReferee.h"


// initialize static data member
wzReferee  wzGame::m_Referee;

///////////////////////////////////////////////////
wzGame::wzGame()
  : m_pBoard        ( new wzBoard() ),
    m_pState        ( wzGameIdleState::Instance() ),
    m_pGameRunner   ( new wzGameRunner( this ) ),
    m_pGameReplayer ( new wzGameReplayer( m_pBoard ) ),
    m_pGameFile     ( new wzGameFile( this ) ),
    m_pSyncObject   ( new wzSyncObject() ),
    m_pPlayerFactory( new wzPlayerFactory() )
{
    m_pGameRunner->SetBlack( m_pPlayerFactory->Create( wzBLACK, m_Options ) );
    m_pGameRunner->SetWhite( m_pPlayerFactory->Create( wzWHITE, m_Options ) );
}

/////////////////////
wzGame::~wzGame() 
{
    delete m_pPlayerFactory;
    delete m_pSyncObject;
    delete m_pGameFile;
    delete m_pGameReplayer;
    delete m_pGameRunner;
    delete m_pBoard;
}

/////////////////////////
void wzGame::NewGame() 
{
    GetState()->NewGame( this );
}

/////////////////////////
void wzGame::StopGame()
{ 
    GetState()->StopGame( this );
}

/////////////////////////
void wzGame::OpenGame()
{ 
    GetState()->OpenGame( this );
}

/////////////////////////
void wzGame::SaveGame()
{ 
    GetState()->SaveGame( this );
}

/////////////////////////
void wzGame::SaveGameAs()
{ 
    GetState()->SaveGameAs( this );
}

/////////////////////////
void wzGame::UndoMove()
{ 
    GetState()->UndoMove( this );
}

/////////////////////////
void wzGame::HintMove()
{ 
    GetState()->HintMove( this );
}

/////////////////////////
void wzGame::FirstPly()
{ 
    GetState()->FirstPly( this );
}

/////////////////////////
void wzGame::PrevPly()
{ 
    GetState()->PrevPly( this );
}

/////////////////////////
void wzGame::NextPly()
{ 
    GetState()->NextPly( this );
}

/////////////////////////
void wzGame::LastPly()
{ 
    GetState()->LastPly( this );
}

/////////////////////////
bool wzGame::SetOptions( const wzOptions& options )
{
    return GetState()->SetOptions( this, options );
}

/////////////////////////
bool wzGame::DoSetOptions( const wzOptions& options )
{
    if( !options.Valid() )
        return false;

    m_Options = options;

    m_pGameRunner->SetBlack( m_pPlayerFactory->Create( wzBLACK, options ) );
    m_pGameRunner->SetWhite( m_pPlayerFactory->Create( wzWHITE, options ) );

    return true;
}

/////////////////////////
// for app to know game state. 
// running means running state or one its substates
////////////////////////////////////////////////////
bool wzGame::IsRunning() const
{
    wzExclusiveZone zone( m_pSyncObject );
    return( m_pState == wzGameRunningState::Instance() || 
            m_pState == wzGameUndoState::Instance() ||
            m_pState == wzGameHintState::Instance() ||
            m_pState == wzGameSaveState::Instance() || 
            m_pState == wzGameSaveAsState::Instance() ); 
}

/////////////////////////
// for player classes to know game state.
// only wzHumanPlayer reacts to HintState.
/////////////////////////////////////////////
bool wzGame::KeepRunning() const
{
    wzExclusiveZone zone( m_pSyncObject );
    return( m_pState == wzGameRunningState::Instance() || 
            m_pState == wzGameHintState::Instance() );
}

/////////////////////////////////////////////
bool wzGame::NeedHint() const
{
    wzExclusiveZone zone( m_pSyncObject );
    return( m_pState == wzGameHintState::Instance() ); 
}

/////////////////////////////////////////////
bool wzGame::NeedUndo() const
{
    wzExclusiveZone zone( m_pSyncObject );
    return( m_pState == wzGameUndoState::Instance() ); 
}

/////////////////////////////////////////////
bool wzGame::NeedSave() const
{
    wzExclusiveZone zone( m_pSyncObject );
    return( m_pState == wzGameSaveState::Instance() ); 
}

/////////////////////////////////////////////
bool wzGame::NeedSaveAs() const
{
    wzExclusiveZone zone( m_pSyncObject );
    return( m_pState == wzGameSaveAsState::Instance() ); 
}

//////////////////////////////////////////////////
std::string wzGame::GetFileName() const
{ 
    return m_pGameFile->GetFileName(); 
}

//////////////////////////////////////////////////
bool  wzGame::IsSaved() const
{ 
    return m_pGameFile->IsSaved(); 
}

////////////////////////////////////////////
void wzGame::Run()
{
    m_pGameRunner->Run();
}

/////////////////////////////////////////////
void wzGame::DoNewGame()
{
    Reset();
    Run();
}

/////////////////////////////////////////////
int wzGame::DoOpenGame()
{
    m_pBoard->DisableView();

    try
    {
        if( !m_pGameFile->Open() )
        {
            m_pBoard->EnableView();
            return FILE_CANCEL;
        }
    }
    catch( const char* pMsg )
    {
        m_pBoard->Clean();
        m_pBoard->EnableView();
        GameFileError( pMsg );
        m_pGameFile->Reset();

        return FILE_ERROR;
    }

    wzSTONE color = (m_pBoard->Plies()%2 == 0)? wzBLACK : wzWHITE;
    m_pGameRunner->CurrentPlayer( color );
    m_pGameReplayer->Reset();
    m_pBoard->EnableView();

    return FILE_OK;
}

/////////////////////////////////////////////
void wzGame::DoSaveGame()
{
    try
    {
        m_pGameFile->Save();
    }
    catch( const char* pMsg )
    {
        GameFileError( pMsg );
    }
}

/////////////////////////////////////////////
void wzGame::DoSaveGameAs()
{
    try
    {
        m_pGameFile->SaveAs();
    }
    catch( const char* pMsg )
    {
        GameFileError( pMsg );
    }
}

/////////////////////////////////////////////
void wzGame::BeSaveState()
{
    ChangeState( wzGameSaveState::Instance() );
}

/////////////////////////////////////////////
void wzGame::BeSaveAsState()
{
    ChangeState( wzGameSaveAsState::Instance() );
}

/////////////////////////////////////////////
void wzGame::BeUndoState()
{
    ChangeState( wzGameUndoState::Instance() );
}

/////////////////////////////////////////////
void wzGame::BeHintState()
{
    ChangeState( wzGameHintState::Instance() );
}

/////////////////////////////////////////////
void wzGame::ResetHint()
{
    wzExclusiveZone zone( m_pSyncObject );
    if( m_pState == wzGameHintState::Instance() )
        m_pState = wzGameRunningState::Instance();
}

/////////////////////////////////////////////
void wzGame::BeEndedState()
{
    ChangeState( wzGameEndedState::Instance() );
    m_pBoard->RemoveHint();
}

/////////////////////////
void wzGame::BeRunningState()
{
    ChangeState( wzGameRunningState::Instance() );
}

/////////////////////////
void wzGame::DoFirstPly()
{
    m_pGameReplayer->FirstPly();
}

/////////////////////////
void wzGame::DoPrevPly()
{
    m_pGameReplayer->PrevPly();
}

/////////////////////////
void wzGame::DoNextPly()
{
    m_pGameReplayer->NextPly();
}

/////////////////////////
void wzGame::DoLastPly()
{
    m_pGameReplayer->LastPly();
}

/////////////////////////
void wzGame::Reset()
{
    m_pBoard->Clean();
    m_pGameRunner->CurrentPlayer( wzBLACK );
    m_pGameReplayer->Reset();
    m_pGameFile->Reset();
    m_Clocks[0].Reset();
    m_Clocks[1].Reset();
}

/////////////////////
void wzGame::SetPlayerFactory(  wzPlayerFactory* pFactory  )
{
    delete m_pPlayerFactory;
    m_pPlayerFactory = pFactory;
}

/////////////////////////
// always access m_pState through GetState() and ChangeState()
//
// GameEndedState is handled within the game loop.
// GameUndoState, GameSaveState and GameSaveAsState are substate
// of GameRunningState, act as flags. No need to notify such 
// changes at the moment.
//////////////////////////////////////////////////////////////
void wzGame::ChangeState( wzGameState* pState )
{
    wzExclusiveZone zone( m_pSyncObject );
    m_pState = pState;

    // notify game state change.
    if( m_pState == wzGameIdleState::Instance() )
        GameIdle();
    else if( m_pState == wzGameRunningState::Instance() )
        GameRunning();
    else if( m_pState == wzGameReplayingState::Instance() )
        GameReplaying();
}

/////////////////////////
// always access m_pState through GetState() and ChangeState()
//////////////////////////////////////////////////////////////
wzGameState* wzGame::GetState()
{
    wzExclusiveZone zone( m_pSyncObject );
    return m_pState;
}

///////////////////////////////
std::string wzGame::GetPlayTime( short iPlayer )
{
    wzExclusiveZone zone( m_pSyncObject );
    std::string stime = m_Clocks[iPlayer].GetTime();

    return stime;
}
