// wzGame.h - The game class
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef __WZ_GAME_H__
#define __WZ_GAME_H__

#include "wzOptions.h"
#include "wzClock.h"
#include <string>

class wzBoard;
class wzReferee;
class wzOptions;
class wzPlayerFactory;
class wzGameRunner;
class wzGameReplayer;
class wzGameState;
class wzGameFile;


//////////////////////////////////
// class represents game
////////////////////////////////////////////////////
class wzGame
{
    friend class wzGameRunner;
    friend class wzGameFile;
    friend class wzGameStateImpl;
    friend class wzGameRunningState;
    friend class wzGameRunningStateImpl;
    friend class wzGameRunningSubStateImpl;
    friend class wzGameEndedState;
    friend class wzGameReplayingState;

    enum { FILE_OK, FILE_CANCEL, FILE_ERROR };

public:
    wzGame();
    virtual ~wzGame();

    // inquiries
    bool  IsRunning()   const;
    bool  KeepRunning() const;
    bool  IsSaved()     const;

    // hint
    bool  NeedHint()    const;
    void  ResetHint();

    wzOptions    GetOptions()  const;
    std::string  GetFileName() const;
    std::string  GetPlayTime( short iPlayer );

    // these delegate to state classes.
    void  NewGame ();
    void  OpenGame();
    void  SaveGame();
    void  SaveGameAs();
    void  StopGame();
    void  UndoMove();
    void  HintMove();
    void  FirstPly();
    void  PrevPly ();
    void  NextPly ();
    void  LastPly ();
    bool  SetOptions( const wzOptions& options );

    static wzReferee  m_Referee;

protected:
    // these called by state classes actually do the jobs
    virtual void  Run();
    virtual void  BeSaveState  ();
    virtual void  BeSaveAsState();
    virtual void  BeUndoState  ();
    virtual void  BeHintState  ();
    virtual void  BeEndedState ();
            void  DoNewGame ();
            int   DoOpenGame();
            void  DoSaveGame();   
            void  DoSaveGameAs();
            void  DoFirstPly();
            void  DoPrevPly ();
            void  DoNextPly ();
            void  DoLastPly ();
            bool  DoSetOptions( const wzOptions& options );

    // notify game state change
    virtual void  GameIdle()      {}
    virtual void  BlackWins()     {}
    virtual void  WhiteWins()     {}
    virtual void  GameDraw()      {}
    virtual void  GameStopped()   {}
    virtual void  GameRunning()   {}
    virtual void  GameReplaying() {}

    // notify game file I/O errors
    virtual void  GameFileError( const char* msg ) {}

    // notify game file changes: 
    // file name change, saved/unsaved
    virtual void  GameFileChange() {}

    // notify start/finish move.
    virtual void  StartMove ( short iPlayer );
    virtual void  FinishMove( short iPlayer );

    // prompt file for open/save
    virtual std::string  PromptFile( bool bForOpen );

    // sync object base and a NullObject, 
    class wzSyncObject
    {
    public:
        wzSyncObject()          {}
        virtual ~wzSyncObject() {}

        virtual void  Lock()    {}
        virtual void  Unlock()  {}
    };

    // exclusive zone
    class wzExclusiveZone
    {
    public:
        wzExclusiveZone( wzSyncObject *pSyncObj ) 
            : m_pSyncObj( pSyncObj ) { m_pSyncObj->Lock(); }
        virtual ~wzExclusiveZone()   { m_pSyncObj->Unlock(); }

    private:
        wzSyncObject*  m_pSyncObj;
    };

    void  SetSyncObject( wzSyncObject* pSyncObj );
    void  SetPlayerFactory( wzPlayerFactory* pFactory );

protected:
    wzBoard*  m_pBoard;

private:
    void  Reset();

    // for GameRunner
    bool  NeedUndo()   const;
    bool  NeedSave()   const;
    bool  NeedSaveAs() const;
    void  BeRunningState();

    // game states
    void          ChangeState( wzGameState* pState );
    wzGameState*  GetState();

private:
    wzOptions         m_Options;
    wzClock           m_Clocks[2];
    wzGameState*      m_pState;
    wzGameRunner*     m_pGameRunner;
    wzGameReplayer*   m_pGameReplayer;
    wzGameFile*       m_pGameFile;
    wzSyncObject*     m_pSyncObject;
    wzPlayerFactory*  m_pPlayerFactory;
};


/////////////////////////////////////////
inline wzOptions wzGame::GetOptions() const
{
    return m_Options;
}

//////////////////////////////////////////////////
inline std::string wzGame::PromptFile( bool bOpen )  
{ 
    return std::string( "untitled.tick5" ); 
}

//////////////////////////////////////////////////
inline void wzGame::SetSyncObject( wzSyncObject* pSyncObj )
{
    delete m_pSyncObject;
    m_pSyncObject = pSyncObj;
}

////////////////////////////////////////
inline void wzGame::StartMove( short iPlayer )
{
    wzExclusiveZone zone( m_pSyncObject );
    m_Clocks[iPlayer].Start();
}

//////////////////////////////////////////
inline void wzGame::FinishMove( short iPlayer )
{
    wzExclusiveZone zone( m_pSyncObject );
    m_Clocks[iPlayer].Stop();
}

#endif
