// wzGameFile.cpp - The game file
// Copyright (C) 2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// Game File format
// ----------------
// byte[2] : ID, 'W' and 'Z'
// byte[3] : version #, major, minor and micro 
// byte[1] : player code, black stone
// byte[?] : player name/Lua script name, delimiter '\n', black stone
// byte[1] : player code, white stone
// byte[?] : player name/Lua script name, delimiter '\n', white stone
// byte[4] : playing time, black stone
// byte[4] : playing time, white stone
// byte[1] : # of moves, 0 to 225
// byte[2] : x- and y-coordinate of a move  |
//   :                                      |
//   :                                       > match # of moves
//   :                                      |
// byte[2] : x- and y-coordinate of a move  |
// byte[1] : trailer, 'Q'


#include "wzGameFile.h"
#include "wzGame.h"
#include "wzBoard.h"
#include "wzVersion.h"
#include "wzMove.h"

#include <fstream>
using namespace std;


////////////////////////////////////////////
const char       wzGameFile::m_MagicID[2] = { 'W', 'Z' };
const char       wzGameFile::m_Trailer = 'Q';
const wzVersion  wzGameFile::m_Version;

////////////////////////////////////////////
wzGameFile::wzGameFile( wzGame* pGame )
  : m_pGame( pGame),
    m_pBoard( pGame->m_pBoard ),
    m_Saved( true )
{
    m_pBoard->Register( this );
}

////////////////////////////////////
void wzGameFile::Reset() 
{
    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    m_FileName.clear();
    m_Saved = true; 
    m_pGame->GameFileChange();
}

////////////////////////////////////////////
bool wzGameFile::Open()
{
    bool bForOpen = true;
    if( !PromptFile( bForOpen ) )
        return false;

    ifstream iFile( m_FileName.c_str(), ios::in|ios::binary );
    if( !iFile )
        throw "Failed to open the game.";

    try
    {
        Read( iFile );
        BeSaved();
    }
    catch( const char* msg )
    {
        throw;
    }
    catch( ... )
    {
        throw "Error in reading the game.";
    }

    return true;
}

////////////////////////////////////////////
void wzGameFile::Read( ifstream& iFile )
{
    iFile.exceptions( ios::failbit|ios::badbit );

    char  MagicID[2];
    iFile.read( MagicID, 2 );
    if( !ValidID( MagicID ) )
        throw "Not a valid Tick5 game file.";

    wzUInt8 VersionNumbers[3];
    iFile.read( (char*)VersionNumbers, 3 );
    // verify version here.

    wzOptions options;
    iFile >> options;
    if( !m_pGame->DoSetOptions( options ) )
        throw "Corrupted file: invalid game options.";

    ReadTime( iFile );

    ApplyMoves( iFile );

    char Trailer;
    iFile >> Trailer;
    if( !ValidTrailer( Trailer ) ) 
        throw "Corrupted file: invalid trailer.";
}
 
////////////////////////////////////////////
void wzGameFile::ReadTime( std::ifstream& iFile )
{
    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    iFile >> m_pGame->m_Clocks[0];
    iFile >> m_pGame->m_Clocks[1];
}

////////////////////////////////////////////
void wzGameFile::ApplyMoves( ifstream& iFile )
{
    m_pBoard->Clean();

    short  nPlies = static_cast<short>( iFile.get() );
    for( short iPly = 1; iPly <= nPlies; iPly++ )
    {
        short  x = static_cast<short>( iFile.get() );
        short  y = static_cast<short>( iFile.get() );

        if( x < 0 || x >= wzDimension || y < 0 || y >= wzDimension )
            throw "Corrupted file: invalid move.";

        wzSTONE stone = (iPly%2 == 1)? wzBLACK: wzWHITE;
        if( !m_pBoard->Set( x, y, stone ) )
            throw "Corrupted file: invalid move.";
    }
}

////////////////////////////////////////////
void wzGameFile::Save()
{
    if( m_FileName.empty() )
        SaveAs();
    else if( !IsSaved() )
        DoSave();
}

////////////////////////////////////////////
void wzGameFile::SaveAs()
{
    bool bForOpen = false;
    if( PromptFile( bForOpen ) )
        DoSave();
}

////////////////////////////////////////////
void wzGameFile::DoSave()
{
    ofstream oFile( m_FileName.c_str(), ios::out|ios::binary|ios::trunc );
    if( !oFile )
        throw "Failed to save the game.";

    try
    {
        Write( oFile );
        BeSaved();
    }
    catch( ... )
    {
        throw "Failed to save the game.";
    }
}

////////////////////////////////////////////
void wzGameFile::Write( std::ofstream& oFile )
{
    oFile.exceptions( ios::failbit|ios::badbit );

    oFile.write( m_MagicID, 2 );

    oFile << m_Version;

    wzOptions options = m_pGame->GetOptions();
    oFile << options;

    WriteTime( oFile );

    WriteMoves( oFile );

    oFile << m_Trailer;
}
 
////////////////////////////////////////////
void wzGameFile::WriteTime( std::ofstream& oFile )
{
    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    oFile << m_pGame->m_Clocks[0];
    oFile << m_pGame->m_Clocks[1];
}
 
////////////////////////////////////////////
void wzGameFile::WriteMoves( std::ofstream& oFile )
{
    wzMoves  Moves;
    m_pBoard->CopyMoves( Moves );
    oFile << static_cast<wzUInt8>( Moves.size() );
    for( wzMovesIterator it = Moves.begin(); it != Moves.end(); it++ )
    {
        oFile << static_cast<wzUInt8>( (*it).X );
        oFile << static_cast<wzUInt8>( (*it).Y );
    }
}

////////////////////////////////////////////
bool wzGameFile::PromptFile( bool bForOpen )
{
    string FileName = m_pGame->PromptFile( bForOpen );
    if( FileName.empty() )
        return false;

    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    m_FileName = FileName;
    m_pGame->GameFileChange();

    return true;
}

////////////////////////////////////////////
void wzGameFile::BoardChanged()
{ 
    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    m_Saved = false; 
    m_pGame->GameFileChange();
} 

////////////////////////////////////////////
void wzGameFile::BeSaved()
{
    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    m_Saved = true; 
    m_pGame->GameFileChange();
}

////////////////////////////////////////////
bool wzGameFile::IsSaved() const
{ 
    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    return m_Saved; 
}

////////////////////////////////////////////
std::string wzGameFile::GetFileName() const
{
    wzGame::wzExclusiveZone zone( m_pGame->m_pSyncObject );
    return m_FileName;
}
