// wzPatternIterator.cpp - Iterator for wzPattern
// Copyright (C) 2006  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzPatternIterator.h"


wzPatternIterator::wzPatternIterator( wzPattern* pBitPattern )
  : m_xStart( 0 ), m_xEnd( wzDimension-1 ), 
    m_yStart( 0 ), m_yEnd( wzDimension-1 ),
    m_Width ( wzDimension ), 
    m_Height( wzDimension ),
    m_x( m_xStart ), m_y( m_yStart ),
    m_pBitPattern( pBitPattern ),
    m_pNextFunc( &wzPatternIterator::NextRectangular )
{
}

bool wzPatternIterator::SetBound( wzBound& bound )
{
    m_xStart = bound.xMin;
    m_yStart = bound.yMin;
    m_xEnd = bound.xMax;
    m_yEnd = bound.yMax;

    m_Width  = m_xEnd - m_xStart + 1;
    m_Height = m_yEnd - m_yStart + 1;

    m_x = m_xStart;
    m_y = m_yStart;

    m_pNextFunc = &wzPatternIterator::NextRectangular;

    return true;
}

void wzPatternIterator::Reset()
{
    m_x = m_xStart;
    m_y = m_yStart;
}

// traverse the range
bool wzPatternIterator::Next()
{
    return (this->*m_pNextFunc)();
}

// traverse a rectangular range
bool wzPatternIterator::NextRectangular()
{
    // loop over the scan range
    m_x++;
    if( m_x > m_xEnd )
    {
        m_x = m_xStart;

        m_y++;
        if( m_y > m_yEnd ) 
           return false;
    }

    return true;
}

// traverse diagonally forward
bool wzPatternIterator::NextDiagonalF()
{
    m_x++;
    m_y++;
    if( m_x > m_xEnd || m_y > m_yEnd )
        return false;
    else
        return true;
}

// traverse diagonally back
bool wzPatternIterator::NextDiagonalB()
{
    m_x--;
    m_y++;
    if( m_x < m_xEnd || m_y > m_yEnd )
        return false;
    else
        return true;
}

wzHorizontal5Iterator::wzHorizontal5Iterator( wzPattern* pBitPattern )
  : wzPatternIterator( pBitPattern )
{ 
    // end of horizontal traverse range
    m_xEnd = m_xEnd - 4; 
}

bool wzHorizontal5Iterator::SetBound( wzBound& bound )
{
    wzPatternIterator::SetBound( bound );

    // end of horizontal traverse range
    m_xEnd = m_xEnd - 4; 

    // the traverse range must wild enough
    return( m_Width >= 5 ); 
}

bool wzHorizontal5Iterator::SetCenter( short x, short y )
{
    // extend 4 both left and right 
    wzBound box;
    box.xMin = x - 4;
    box.yMin = y;
    box.xMax = x + 4;
    box.yMax = y;

    // must be with the board
    if( box.xMin < 0 )  box.xMin = 0;
    if( box.xMax >= wzDimension )  box.xMax = wzDimension - 1;

    return SetBound( box );
}

// get start and end positions of a 5-in-a-row
wzBound wzHorizontal5Iterator::GetRow() const
{
    wzBound bound;

    // start position
    bound.xMin = m_x;
    bound.yMin = m_y;

    // end position
    bound.xMax = m_x + 4;
    bound.yMax = m_y;

    return bound;
}

wzVertical5Iterator::wzVertical5Iterator( wzPattern* pBitPattern )
  : wzPatternIterator( pBitPattern )
{ 
    // end of vertical traverse range
    m_yEnd = m_yEnd - 4;
}

bool wzVertical5Iterator::SetBound( wzBound& bound )
{
    wzPatternIterator::SetBound( bound );

    // end of vertical traverse range
    m_yEnd = m_yEnd - 4;

    // traverse range must be tall enough
    return( m_Height >= 5 ); 
}

bool wzVertical5Iterator::SetCenter( short x, short y )
{
    // extend 4 both up and down 
    wzBound box;
    box.xMin = x;
    box.yMin = y - 4;
    box.xMax = x;
    box.yMax = y + 4;

    // still within the board
    if( box.yMin < 0 )  box.yMin = 0;
    if( box.yMax >= wzDimension )  box.yMax = wzDimension - 1;

    return SetBound( box );
}

// get 1st and last positions of a 5-in-a-row
wzBound wzVertical5Iterator::GetRow() const
{
    wzBound bound;

    // start position
    bound.xMin = m_x;
    bound.yMin = m_y;

    // end position
    bound.yMax = m_y + 4;
    bound.xMax = m_x;

    return bound;
}

wzDiagonal5FIterator::wzDiagonal5FIterator( wzPattern* pBitPattern )
  : wzPatternIterator( pBitPattern )
{ 
    // ends of traverse range
    m_xEnd = m_xEnd - 4; 
    m_yEnd = m_yEnd - 4;
}

bool wzDiagonal5FIterator::SetBound( wzBound& bound )
{
    wzPatternIterator::SetBound( bound );

    // ends of traverse range
    m_xEnd = m_xEnd - 4; 
    m_yEnd = m_yEnd - 4;

    // traverse range must be wild and tall enough
    return( m_Width >= 5 && m_Height >= 5 ); 
}

bool wzDiagonal5FIterator::SetCenter( short x, short y )
{
    m_xStart = x;
    m_yStart = y;
    m_xEnd   = x;
    m_yEnd   = y;

    short i = 0;
    while( i < 4 && m_xStart > 0 && m_yStart > 0  )
    {
        m_xStart--;
        m_yStart--;
        i++;
    } 

    i = 0;
    while( i < 4 && m_xEnd < wzDimension-1 && m_yEnd < wzDimension-1 )
    {
        m_xEnd++;
        m_yEnd++;
        i++;
    }

    m_Width  = m_xEnd - m_xStart + 1;
    m_Height = m_yEnd - m_yStart + 1;

    m_xEnd = m_xEnd - 4; 
    m_yEnd = m_yEnd - 4;

    m_x = m_xStart;
    m_y = m_yStart;

    // pick a function for Next()
    //   this causes compilation error
    //   m_pNextFunc = &wzPatternIterator::NextDiagonalF;
    m_pNextFunc = &wzDiagonal5FIterator::NextDiagonalF;

    // traverse range must be wild and tall enough
    return( m_Width >= 5 && m_Height >= 5 ); 
}

// get 1st and last positions of a 5-in-a-row
wzBound wzDiagonal5FIterator::GetRow() const
{
    wzBound bound;

    // start position
    bound.xMin = m_x;
    bound.yMin = m_y;

    // end position
    bound.xMax = m_x + 4;
    bound.yMax = m_y + 4;

    return bound;
}

wzDiagonal5BIterator::wzDiagonal5BIterator( wzPattern* pBitPattern )
  : wzPatternIterator( pBitPattern )
{ 
    //Init();

    // ends of traverse range
    m_xEnd = m_xEnd - 4; 
    m_yEnd = m_yEnd - 4;
}

bool wzDiagonal5BIterator::SetBound( wzBound& bound )
{
    wzPatternIterator::SetBound( bound );

    // ends of traverse range
    m_xEnd = m_xEnd - 4; 
    m_yEnd = m_yEnd - 4;

    // traverse range must be wild and tall enough
    return( m_Width >= 5 && m_Height >= 5 ); 
}

bool wzDiagonal5BIterator::SetCenter( short x, short y )
{
    m_xStart = x;
    m_yStart = y;
    m_xEnd   = x;
    m_yEnd   = y;

    // right most point
    short i = 0;
    while( i < 4 && m_xStart < wzDimension-1 && m_yStart > 0 )
    {
        m_xStart++;
        m_yStart--;
        i++;
    } 

    // left most point
    i = 0;
    while( i < 4 && m_xEnd > 0 && m_yEnd < wzDimension-1 )
    {
        m_xEnd--;
        m_yEnd++;
        i++;
    }

    m_Width  = m_xStart - m_xEnd + 1;
    m_Height = m_yEnd - m_yStart + 1;

    m_xStart = m_xStart - 4;
    m_yEnd   = m_yEnd - 4;

    m_x = m_xStart;
    m_y = m_yStart;

    // pick a cuntction for Next()
    m_pNextFunc = &wzDiagonal5BIterator::NextDiagonalB;

    // traverse range must be wild and tall enough
    return( m_Width >= 5 && m_Height >= 5 ); 
}

// get 1st and last positions of a 5-in-a-row
wzBound wzDiagonal5BIterator::GetRow() const
{
    wzBound bound;

    // start position
    bound.xMin = m_x;
    bound.yMin = m_y + 4;

    // end position
    bound.xMax = m_x + 4;
    bound.yMax = m_y;

    return bound;
}
