// wzLuaBoard.cpp - Board for Lua player
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzLuaBoard.h"
#include <wzBoard.h>
#include <wzMoveObserver.h>

//#include <sstream>


const char wzLuaBoard::m_ClassName[] = "WZ_LUA_BOARD";


//////////////////////////////////////////////
wzLuaBoard::wzLuaBoard( const wzBoard& board )
  : wzBoardBase( board ),
    m_pMoveObserver( new wzMoveObserver() )
{
    wzBoardBase::Register( m_pMoveObserver );
    wzMoves moves;
    board.CopyMoves( moves );
    m_pMoveObserver->SetMoves( moves );
}

//////////////////////////////////////
wzLuaBoard::wzLuaBoard( const wzLuaBoard& other )
  : wzBoardBase( other ),
    m_pMoveObserver( new wzMoveObserver() )
{
    wzBoardBase::Register( m_pMoveObserver );
    wzMoves moves;
    other.m_pMoveObserver->CopyMoves( moves );
    m_pMoveObserver->SetMoves( moves );
}

////////////////////////
wzLuaBoard::~wzLuaBoard()
{
    delete m_pMoveObserver;
}

///////////////////
// Names of the functions to be exposed
//////////////////////////////////////////////////
#define METHOD(class, name) {#name, class::name}
const luaL_reg wzLuaBoard::m_Methods[] = 
{
    METHOD( wzLuaBoard, isempty ),
    METHOD( wzLuaBoard, isfull ),
    METHOD( wzLuaBoard, inboard ),
    METHOD( wzLuaBoard, getplies ),
    METHOD( wzLuaBoard, getstone ),
    METHOD( wzLuaBoard, dimension ),
    METHOD( wzLuaBoard, lastmove ),
    METHOD( wzLuaBoard, putstone ),
    METHOD( wzLuaBoard, remove ),
    METHOD( wzLuaBoard, clean ),
    METHOD( wzLuaBoard, clone ),

    // deprecated
    { "set",   wzLuaBoard::putstone },
    { "reset", wzLuaBoard::remove },

    { NULL, NULL }
};

///////////////////////////////////
// Register the object
// This function actually creates a metatable with both __metatable
// and __index point to another table that contains all the pointers
// of the methods listed in m_Methods. 
/////////////////////////////////////////////////////////////////
void wzLuaBoard::Register( lua_State* pLS ) 
{
    // create a table and fill it up with methods
    lua_newtable( pLS );                 
    int methodtable = lua_gettop( pLS );
    luaL_openlib( pLS, NULL, m_Methods, 0 );
    
    // create a metatable using the class name
    luaL_newmetatable( pLS, m_ClassName ); 
    int metatable = lua_gettop( pLS );
    
    // metatable field __metatable pointing to methodtable 
    lua_pushliteral( pLS, "__metatable" );
    lua_pushvalue( pLS, methodtable );
    lua_settable( pLS, metatable ); 
     
    // metatable field __index pointing to methodtable 
    lua_pushliteral( pLS, "__index" );
    lua_pushvalue( pLS, methodtable );
    lua_settable( pLS, metatable );
    
    // metatable field __gc pointing to garbage collector
    lua_pushliteral( pLS, "__gc");
    lua_pushcfunction( pLS, Destroy );
    lua_settable( pLS, metatable );

    // clean up: drop both tables
    lua_pop( pLS, 2 );  
}

///////////////////
// cast the C++ object into a Lua object
/////////////////////////////////////////////////////////////
int wzLuaBoard::Cast2Lua( lua_State* pLS, wzLuaBoard* pBoard )
{
    // create userdata
    lua_boxpointer( pLS, pBoard );

    // associate the metatble with the userdata
    luaL_getmetatable( pLS, m_ClassName );
    lua_setmetatable ( pLS, -2 );

    // return the new object
    return 1;
}

///////////////////
// Retrieve the C++ object from its Lua object.
// This function will not return if it cannot get the pointer, 
// because it will call luaL_typerror() which never returns.
/////////////////////////////////////////////////////////////
wzLuaBoard* wzLuaBoard::ValidBoard( lua_State* pLS, const int& iArg ) 
{
    void* pUD = luaL_checkudata( pLS, iArg, m_ClassName );
    if( pUD == NULL ) 
        luaL_typerror( pLS, iArg, "board" );
    
    wzLuaBoard* pBoard = *( static_cast<wzLuaBoard**>( pUD ) );

    return pBoard; 
}

///////////////////
// Retrieve argument for x or y coordinates and validate it. 
// Never return if it is invalid.
///////////////////////////////////////////////////////////
int wzLuaBoard::ValidXY( lua_State* pLS, const int& iArg )
{
    int xy = luaL_checkint( pLS, iArg );
    if( !wzBoardBase::Valid( xy ) )
    {
#if 0
        // valgrind reported leaks by << operator. valgrind-3.2.1
        // still complained a leak by str(), after it is put in the
        // block; while valgrind-2.4.0 didn't.  
        std::string  msg;
        {
            std::ostringstream  strs;
            strs << "number [0, " << wzDimension - 1 << "] expected, got " << xy;
            msg = strs.str();
        }

        luaL_argerror( pLS, iArg, msg.c_str() );
#else
        char  msg[80];
        snprintf( msg, 80, "number within [0, %d] expected, got %d", wzDimension-1, xy );
        luaL_argerror( pLS, iArg, msg );
#endif
    }

    return xy;
}

///////////////////
// Retrieve argument for color and validate it. 
// Never return if it is invalid.
///////////////////////////////////////////////////////////
wzSTONE wzLuaBoard::ValidColor( lua_State* pLS, const int& iArg )
{
    int color = luaL_checkint( pLS, iArg );

    wzSTONE stone = wzNONE;
    if( color == 1 )
    {
        stone = wzBLACK;
    }
    else if( color == -1 )
    {
        stone = wzWHITE;
    }
    else
    {
#if 0
        // see comments in ValidXY()
        std::string  msg;
        {
            std::ostringstream  strs;
            strs << "1(for black) or -1(for white) expected, got " << color;
            msg = strs.str();
        }

        luaL_argerror( pLS, iArg, msg.c_str() );
#else
        char  msg[80];
        snprintf( msg, 80, "1(for black) or -1(for white) expected, got %d", color );
        luaL_argerror( pLS, iArg, msg );
#endif
    }

    return stone;
}

////////////////////////////
// garbage collector
///////////////////////////////////////////
int wzLuaBoard::Destroy( lua_State* pLS )
{
    // verify 1st arg is userdata of wzLuaBoard object
    void* pUD = luaL_checkudata( pLS, 1, m_ClassName );
    if( pUD == NULL ) 
        return 0;

    wzLuaBoard* pBoard = *( static_cast<wzLuaBoard**>( pUD ) );
    if( pBoard != NULL )
    {
        delete pBoard;

        // indicate the object deleted
        lua_unboxpointer( pLS, 1) = NULL;
    }

    return 0;
}

///////////////////
// board:isfull()
////////////////////////////////////////////
int wzLuaBoard::isfull( lua_State* pLS )
{
    bool bRet = ValidBoard( pLS, 1 )->IsFull();
    lua_pushboolean( pLS, bRet );

    return 1;
}

///////////////////
// board:isempty() or board:isempty( x, y )
////////////////////////////////////////////
int wzLuaBoard::isempty( lua_State* pLS )
{
    bool bRet = false;
    wzLuaBoard* pBoard = ValidBoard( pLS, 1 );

    // number of arguments
    int nArgs = lua_gettop( pLS );

    // board:isempty()
    if( nArgs == 1 )
    {
        bRet = pBoard->IsEmpty();
    }
    else // board:isempty( x, y )
    {
        int x = ValidXY( pLS, 2 );
        int y = ValidXY( pLS, 3 );

        bRet = pBoard->IsEmpty( x, y );
    }

    lua_pushboolean( pLS, bRet );
    return 1;
}

///////////////////
// board:inboard( x, y )
////////////////////////////////////////////
int wzLuaBoard::inboard( lua_State* pLS )
{
    ValidBoard( pLS, 1 );

    int x = luaL_checkint( pLS, 2 );
    int y = luaL_checkint( pLS, 3 );

    bool bRet = wzBoardBase::Valid( x ) && wzBoardBase::Valid( y );
    lua_pushboolean( pLS, bRet );

    return 1;
}

///////////////////
// board:getplies()
////////////////////////////////////////////
int wzLuaBoard::getplies( lua_State* pLS )
{
    short nPlies = ValidBoard( pLS, 1 )->Plies();
    lua_pushnumber( pLS, nPlies );

    return 1;
}

///////////////////
// board:getstone( x, y )
////////////////////////////////////////////
int wzLuaBoard::getstone( lua_State* pLS )
{
    int x = ValidXY( pLS, 2 );
    int y = ValidXY( pLS, 3 );

    wzSTONE stone = ValidBoard( pLS, 1 )->GetStone( x, y );
    lua_pushnumber( pLS, stone );

    return 1;
}

///////////////////
// board:dimension()
////////////////////////////////////////////
int wzLuaBoard::dimension( lua_State* pLS )
{
    // verify 1st arg be a board
    ValidBoard( pLS, 1 );

    lua_pushnumber( pLS, wzDimension );
    return 1;
}

///////////////////
// x, y, stone = board:lastmove()
////////////////////////////////////////////
int wzLuaBoard::lastmove( lua_State* pLS )
{
    wzLuaBoard* pBoard = ValidBoard( pLS, 1 );
    if( pBoard->IsEmpty() )
    {
       lua_pushnil( pLS );
       lua_pushnil( pLS );
       lua_pushnil( pLS );
    }
    else
    {
       wzMove move = pBoard->m_pMoveObserver->GetLastMove();
       lua_pushnumber( pLS, move.X );
       lua_pushnumber( pLS, move.Y );
       lua_pushnumber( pLS, move.Stone );
    }

    return 3;
}

///////////////////
// board:putstone( x, y, stone )
// board:set( x, y, stone )
////////////////////////////////////////////
int wzLuaBoard::putstone( lua_State* pLS )
{
    int  x = ValidXY( pLS, 2 );
    int  y = ValidXY( pLS, 3 );

    wzSTONE  stone = ValidColor( pLS, 4 );

    bool bRet = ValidBoard( pLS, 1 )->Set( x, y, stone );

    lua_pushboolean( pLS, bRet );
    return 1;
}

///////////////////
// board:remove( x, y )
// board:reset( x, y )
////////////////////////////////////////////
int wzLuaBoard::remove( lua_State* pLS )
{
    int x = ValidXY( pLS, 2 );
    int y = ValidXY( pLS, 3 );
    
    bool bRet = ValidBoard( pLS, 1 )->Reset( x, y );

    lua_pushboolean( pLS, bRet );
    return 1;
}

///////////////////
// board:clean()
////////////////////////////////////////////
int wzLuaBoard::clean( lua_State* pLS )
{
    ValidBoard( pLS, 1 )->Clean();

    return 0;
}

///////////////////
// board2 = board:clone()
//////////////////////////////////////////
int wzLuaBoard::clone( lua_State* pLS ) 
{
    wzLuaBoard* pOld = ValidBoard( pLS, 1 );
    wzLuaBoard* pNew = pOld->Clone();

    return Cast2Lua( pLS, pNew );
}
