// wzLuaPlayer.cpp - Player of AI built using Lua
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "wzLuaPlayer.h"
#include "wzLuaVM.h"
#include "wzLuaBoard.h"
#include <wzBoard.h>
#include <wzGame.h>

#include <sstream>


const char wzLuaPlayer::m_ClassName[]     = "WZ_LUA_PLAYER";
const char wzLuaPlayer::m_GamePointer[]   = "WZ_GAME_POINTER";
const char wzLuaPlayer::m_LuaAIFunction[] = "tick5think";


/////////////////////////////////////////////////////////////
wzLuaPlayer::wzLuaPlayer( wzSTONE stone, const std::string& LuaScript, wzLuaVM* pLuaVM )
  : wzPlayer( stone ),
    m_pLuaVM( pLuaVM ),
    m_LuaScript( LuaScript )
{
    m_pLuaVM->Register( wzLuaBoard::Register );
    m_pLuaVM->Register( wzLuaPlayer::Register );
    m_pLuaVM->RegPointer( m_ClassName, this );
}

//////////////////////
wzLuaPlayer::~wzLuaPlayer()
{
    delete m_pLuaVM;
}

////////////////////////////////////////////////////////
void wzLuaPlayer::Init()
{
    if( !m_pLuaVM->RunScript( m_LuaScript.c_str() ) )
        m_pGame->StopGame();
}

////////////////////////////////////////////////////////
void wzLuaPlayer::Set( wzGame* pGame, wzBoard* pBoard )
{
    wzPlayer::Set( pGame, pBoard );

    m_pLuaVM->RegPointer( m_GamePointer, m_pGame );
}

////////////////////////////////////
bool wzLuaPlayer::MakeAMove( wzMove& move )
{
    if( !m_pGame->KeepRunning() )
        return false;

    wzLuaBoard* pLuaBoard = new wzLuaBoard( *m_pBoard );
    int  x, y;
    if( LuaAIMakeAMove( pLuaBoard, x, y ) )
    {
        move.X = static_cast<short>( x );
        move.Y = static_cast<short>( y );
        move.Stone = m_Stone;
        return true;
    }
    else // something wrong with the Lua script,
    {
        m_pGame->StopGame();
        return false;
    }
}

////////////////////////////////////////////////////////////////////////
// Call Lua AI function: tick5think( board )
//    Return true if a valid move is generated; false, if any 
//    of the following errors occurs during the call. 
//    Errors:
//      - Lua function tick5think() is not found
//      - Error in the results returned by Lua AI function:
//        (lua_pcall() returns 0)
//          1. returns none or one values
//          2. returns non-numeric values
//          3. returns coordinates that exceed boundary
//          4. returns coordinates of an occupided spot,
//             i.e., attempt a invalid move.        
//      - Call to Lua AI function failed: 
//        (lua_pcall() returns LUA-ERRRUN, LUA_ERRMEM or LUA_ERRERR )
//          1. directly called error()
//          2. errors occured in any of the API functions
//          3. etc. 
//////////////////////////////////////////////////////////////////////////
bool wzLuaPlayer::LuaAIMakeAMove( wzLuaBoard* pBoard, int& x, int& y )
{
    bool bRet = true;
    if( LoadLuaAIFunc( m_pLuaVM->m_pLS ) ) 
    {
        if( CallLuaAIFunc( m_pLuaVM->m_pLS, pBoard ) )
        {
            bRet = Retrieve( x, y );
        }
        else 
        {
            // call failed, flush whatever 
            // on top of the stack
            m_pLuaVM->CallAlert();

            bRet = false;
        }
    } 
    else
    {
        lua_pop( m_pLuaVM->m_pLS, 1 );
        LuaAIError();

        // hasn't been casted into Lua, 
        // not subject to garbage collection
        delete pBoard;

        bRet = false;
    }

    return bRet;
}

//////////////////////////////////////////////////
bool wzLuaPlayer::LoadLuaAIFunc( lua_State* pLS )
{
    lua_getglobal( pLS, m_LuaAIFunction ); 
    return lua_isfunction( pLS, -1 );
}

////////////////////////////////////////////////////////////////////
bool wzLuaPlayer::CallLuaAIFunc( lua_State* pLS, wzLuaBoard* pBoard )
{
    int handler = LoadErrHandler( pLS );
    wzLuaBoard::Cast2Lua( pLS, pBoard );

    // Call Lua AI function: 
    //   1 arg(board), 2 results, 1 err handler(if any)
    return( lua_pcall( pLS, 1, 2, handler ) == 0 ); 
}

///////////////////////////////
// load error handler (_TRACEBACK) and insert it 
// under Lua AI function(just loaded and on stack top)
// Return: 0, if failded to load;
//         index to the handler, otherwise;
///////////////////////////////////////////////////////
int wzLuaPlayer::LoadErrHandler( lua_State* pLS )
{
    int pos = lua_gettop( pLS );
    lua_getglobal( pLS, "_TRACEBACK" );

    if( lua_isfunction( pLS, -1 ) )
    {
        lua_insert( pLS, pos );
        return pos;
    }
    else
    {
        lua_pop( pLS, 1 );
        return 0;
    }
}

///////////////////
// Retrieve results returned by Lua AI function
////////////////////////////////////////////
bool wzLuaPlayer::Retrieve( int& x, int& y )
{
    std::string msg;
    bool bRet = true;

    int xType = lua_type( m_pLuaVM->m_pLS, -2 ); 
    int yType = lua_type( m_pLuaVM->m_pLS, -1 );

    if( xType == LUA_TNUMBER && yType == LUA_TNUMBER )
    {
        x = static_cast<int>( lua_tonumber( m_pLuaVM->m_pLS, -2 ) );
        y = static_cast<int>( lua_tonumber( m_pLuaVM->m_pLS, -1 ) );

        lua_pop( m_pLuaVM->m_pLS, 2 );

        if( !Valid( x, y, msg ) )
        {
            LuaAIError( msg );
            bRet = false;
        }
    }
    else
    { 
        lua_pop( m_pLuaVM->m_pLS, 2 );

        ReturnTypeError( xType, yType, msg );
        LuaAIError( msg );
        bRet = false;
    }

    return bRet;
}

///////////////////
// Validate x or y coordinates
/////////////////////////////////////////////////////////////
bool wzLuaPlayer::Valid( const int& x, const int& y, std::string& msg )
{
    if( wzBoardBase::Valid( x ) && wzBoardBase::Valid( y ) )
    {
        if( m_pBoard->IsEmpty( x, y ) )
        {
            return true;
        }
        else
        {
            std::stringstream  strs;
            strs << "returned spot (" << x << ", " << y << ") not empty";
            msg = strs.str();

            return false;
        }
    }

    std::stringstream  strs;
    strs << "return values within [0, " << wzDimension - 1 
         << "] expected, got " << x << " and " << y;
    msg = strs.str();

    return false;
}

////////////////////////////////////////////////////////////////////
void wzLuaPlayer::ReturnTypeError( const int& xType, const int& yType, 
                                   std::string& msg )
{
    std::stringstream  strs;
    strs << "two return numbers expected, got " 
         << lua_typename( m_pLuaVM->m_pLS, xType )
         << " and "
         << lua_typename( m_pLuaVM->m_pLS, yType );
    msg = strs.str();
}

///////////////////
// out put error msg
////////////////////////////////////////////////
void wzLuaPlayer::LuaAIError( std::string& msg )
{
    std::string LongMsg = m_LuaScript;
    LongMsg += ": error in `";
    LongMsg += m_LuaAIFunction;
    LongMsg += "' (";
    LongMsg += msg;
    LongMsg += ")";

    lua_pushstring( m_pLuaVM->m_pLS, LongMsg.c_str() );
    m_pLuaVM->CallAlert();
}

///////////////////
// out put error msg
///////////////////////////////
void wzLuaPlayer::LuaAIError()
{
    std::string LongMsg = m_LuaScript;
    LongMsg += ": function `";
    LongMsg += m_LuaAIFunction;
    LongMsg += "' not found";

    lua_pushstring( m_pLuaVM->m_pLS, LongMsg.c_str() );
    m_pLuaVM->CallAlert();
}

///////////////////
// Register Lua global functions
///////////////////////////////////////////
void wzLuaPlayer::Register( lua_State* pLS )
{
    // register global functions 
    lua_register( pLS, "mycolor",     MyColor );
    lua_register( pLS, "keeprunning", KeepRunning );
}

///////////////////
// implement global function: mycolor() 
////////////////////////////////////////////
int wzLuaPlayer::MyColor( lua_State* pLS )
{
    wzLuaPlayer* pPlayer = static_cast<wzLuaPlayer*>( wzLuaVM::GetPointer( pLS, m_ClassName ) );
    lua_pushnumber( pLS, pPlayer->GetStone() );

    return 1;
}

///////////////////
// implement global function: keeprunning() 
// it's better that Lua script calls this function in loops
/////////////////////////////////////////////////////////////
int wzLuaPlayer::KeepRunning( lua_State* pLS )
{
    bool bRet = false;

    wzGame* pGame = static_cast<wzGame*>( wzLuaVM::GetPointer( pLS, m_GamePointer ) );
    if( pGame != NULL )
        bRet = pGame->KeepRunning();

    lua_pushboolean( pLS, bRet );
    return 1;
}
