// wzGUIBoardView.cpp - Board view for GUI app.
// Copyright (C) 2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif

#include "wzGUIBoardView.h"
#include "wzBoard.h"
#include "wzEvent.h"
#include "wzIDs.h"


BEGIN_EVENT_TABLE ( wzGUIBoardView, wxWindow )
    EVT_PAINT     ( wzGUIBoardView::OnPaint )
    EVT_LEFT_UP   ( wzGUIBoardView::OnClick )
    EVT_WZ_DISPLAY( wzGUIBoardView::OnDisplay )
    EVT_TIMER     ( wzID_HILIGHT_TIMER, wzGUIBoardView::OnHilight )
    EVT_TIMER     ( wzID_HINT_STOPPER,  wzGUIBoardView::OnStopHint )
END_EVENT_TABLE()


////////////////////////////
wzGUIBoardView::wzGUIBoardView( wxWindow* parent )
  : wzBoardView(),
    wxWindow( parent, wxID_ANY ),
    m_HilightTimer( this, wzID_HILIGHT_TIMER ),
    m_HintStopper ( this, wzID_HINT_STOPPER ),
    m_PaintHilight( true )
{
    // Eleminate background update to avoid flicker.
    // OnPaint() paints exactly what it looks like.
    SetBackgroundStyle( wxBG_STYLE_CUSTOM );

    BigView();

    m_pMutex = new wxMutex(),
    m_pCondition = new wxCondition( *m_pMutex );
}

/////////////////////////
wzGUIBoardView::~wzGUIBoardView()
{
    delete m_pCondition;
    delete m_pMutex;
}

///////////////////////////////
short wzGUIBoardView::ToggleView()
{
    if( IsBigView() )
        SmallView();
    else
        BigView();

    Refresh();

    return Width();
}

///////////////////////////////
void wzGUIBoardView::BigView()
{
    LoadBigPads();

    m_PadWidth  = 40;  // pixels
    m_ViewWidth = m_PadWidth*wzDimension;
    m_Half      = m_PadWidth/2;
    m_One3rd    = m_PadWidth/3;
    m_WidePen   = 7;
    m_NarrowPen = 2;
}

///////////////////////////////
void wzGUIBoardView::SmallView()
{
    LoadSmallPads();

    m_PadWidth  = 32;  // pixels
    m_ViewWidth = m_PadWidth*wzDimension;
    m_Half      = m_PadWidth/2;
    m_One3rd    = m_PadWidth/3;
    m_WidePen   = 5;
    m_NarrowPen = 2;
}

/////////////////////////
void wzGUIBoardView::UpdateView()
{
    m_PaintHilight = true;
    if( m_HilightTimer.IsRunning() )
        m_HilightTimer.Stop();

    wzDisplayEvent event;
    wxPostEvent( this, event );
}

/////////////////////////
// Wait for Signal() being called. 
// This function is called by wzHumanPlayer in worker thread.
/////////////////////////////////////////////////////////////
bool wzGUIBoardView::WaitForInput( short& x, short& y)
{
    wxMutexLocker lock( *m_pMutex );
    m_pCondition->Wait();

    x = m_xClick;
    y = m_yClick;

    return true;
}

//////////////////
// Signal a waiting wzHumanPlayer, a mouse clicking happened
////////////////////////////////////////////////////////////
void wzGUIBoardView::NotifyInput()
{
    wxMutexLocker lock( *m_pMutex );
    m_pCondition->Signal();
}

///////////////////////
// Mouse click event handler, called by main thread.
// Signal wzHumanPlayer, if there's one, that mouse clicked.
//////////////////////////////////////////////////////////
void wzGUIBoardView::OnClick( wxMouseEvent& event )
{
    wxCoord x, y;
    event.GetPosition( &x, &y );

    if( WithinView( x, y ) )
    {
        m_xClick = x/m_PadWidth;
        m_yClick = y/m_PadWidth;

        if( m_pBoard->IsEmpty( m_xClick, m_yClick ) )
            NotifyInput();
    }
}

//////////////////////////////
// Selectively refresh view upon the event.
// On wxMSW, it can also reduce flicker when hint is flashing
//////////////////////////////////////////////////////////////
void wzGUIBoardView::OnHilight( wxTimerEvent& event )
{
    if( m_pBoard->HasWinner()  )
    {
        RefreshWinningRow();
    }
    else 
    {
        if( m_pBoard->HasHint() )
            RefreshHint();
        else if( !m_pBoard->IsEmpty() )
            RefreshLastMove();
    }
}

///////////////////////////////////////
void wzGUIBoardView::RefreshWinningRow()
{
    wzBound WinningRow = m_pBoard->GetWinningRow();
    short  xmin = (WinningRow.xMin <= WinningRow.xMax)? WinningRow.xMin: WinningRow.xMax;
    short  ymin = (WinningRow.yMin <= WinningRow.yMax)? WinningRow.yMin: WinningRow.yMax;
    short  width  = (abs(WinningRow.xMax - WinningRow.xMin) + 1)*m_PadWidth;
    short  height = (abs(WinningRow.yMax - WinningRow.yMin) + 1)*m_PadWidth;

    RefreshRect( wxRect( xmin*m_PadWidth, ymin*m_PadWidth, width, height ) );
}

//////////////////////////////////////
void wzGUIBoardView::RefreshLastMove()
{
    wzMove move = m_pBoard->GetLastMove();
    short  x = move.X*m_PadWidth;
    short  y = move.Y*m_PadWidth;

    wxRect rect( x, y, m_PadWidth, m_PadWidth );
    // bigger to fully cover the circle.
    rect.Inflate( m_One3rd, m_One3rd );

    RefreshRect( rect );
}

//////////////////////////////////
void wzGUIBoardView::RefreshHint()
{
    wzMove move = m_pBoard->GetHint();
    short  x = move.X*m_PadWidth;
    short  y = move.Y*m_PadWidth;

    RefreshRect( wxRect( x, y, m_PadWidth, m_PadWidth ) );
}

////////////////////////////////////////////////////
void wzGUIBoardView::OnStopHint( wxTimerEvent& event )
{
    // Stopping timer won't affect painting last move (if any),
    // because wzBoard::RemoveHint() will update the view;
    // it eleminates randomness happened when painting last move 
    // continues.
    if( m_HilightTimer.IsRunning() )
        m_HilightTimer.Stop();

    m_pBoard->RemoveHint();
}

//////////////////////////////////////////////////
void wzGUIBoardView::OnPaint( wxPaintEvent& event )
{
    wxPaintDC dc( this );
    PaintBoard( dc );

    // selectively paint hilights
    if( m_pBoard->HasWinner()  )
    {
        BlinkWinningRow( dc );
    }
    else 
    {
        if( m_pBoard->HasHint() )
        {
            BlinkHint( dc );

            // statically paint last move, if any
            if( !m_pBoard->IsEmpty() )
                HiglightLastMove( dc );
        }
        else if( !m_pBoard->IsEmpty() )
        {
            BlinkLastMove( dc );
        }
    }
}

///////////////////////////////////////////
void wzGUIBoardView::BlinkWinningRow( wxDC& dc )
{
    if( TimeToPaint() )
        HiglightWinningRow( dc );

    KeepBlinking( ONE_SECOND );
}

///////////////////////////////////////////
void wzGUIBoardView::BlinkLastMove( wxDC& dc )
{
    if( TimeToPaint() )
        HiglightLastMove( dc );

    KeepBlinking( ONE_SECOND );
}

////////////////////////////////////////
void wzGUIBoardView::BlinkHint( wxDC& dc )
{
    // 1st time call this function
    if( !m_HintStopper.IsRunning() )
    {
        m_HintStopper.Start( FIVE_SECONDS, wxTIMER_ONE_SHOT );

        m_PaintHilight = true;

        if( m_HilightTimer.IsRunning() )
            m_HilightTimer.Stop();
    }

    if( TimeToPaint() )
        HiglightHint( dc );

    KeepBlinking( HALF_SECOND );
}

////////////////////
// Toggle the flag, making the hilight alternatively 
// be painted, i.e., blinking.
/////////////////////////////////////////////////
bool wzGUIBoardView::TimeToPaint()
{
    bool bRet = m_PaintHilight;

    // toggle the flag
    m_PaintHilight = !m_PaintHilight;

    return bRet;
}

////////////////////
// One-shot timer, to avoid unnecessary timer event,
// e.g., Undo Moves to empty board.
////////////////////////////////////////////////////
void wzGUIBoardView::KeepBlinking( const int& millisecs )
{
    if( !m_HilightTimer.IsRunning() )
        m_HilightTimer.Start( millisecs, wxTIMER_ONE_SHOT );
}

////////////////////////////////////////////////////
void wzGUIBoardView::PaintBoard( wxDC& dc )
{
    wxMemoryDC mdc;

    // top-left
    short  x = 0;
    short  y = 0;
    PaintPad( dc, mdc, x, y, m_PadsTopLeft );

    // top-middle
    for( x = 1; x < wzDimension-1; ++x )
        PaintPad( dc, mdc, x, y, m_PadsTopMiddle );

    // top-right
    x = wzDimension - 1;
    PaintPad( dc, mdc, x, y, m_PadsTopRight );

    // central lines
    for( y = 1; y < wzDimension - 1; ++y )
    {
        // left edge
        x = 0;
        PaintPad( dc, mdc, x, y, m_PadsCentralLeft );

        // middle
        for( x = 1; x < wzDimension-1; ++x )
            PaintPad( dc, mdc, x, y, m_PadsCentralMiddle );

        // right edge
        x = wzDimension - 1;
        PaintPad( dc, mdc, x, y, m_PadsCentralRight );
    }

    // bottom-left
    y = wzDimension - 1;
    x = 0;
    PaintPad( dc, mdc, x, y, m_PadsBottomLeft );

    // bottom-middle
    for( x = 1; x < wzDimension-1; ++x )
        PaintPad( dc, mdc, x, y, m_PadsBottomMiddle );

    // bottom-right
    x = wzDimension - 1;
    PaintPad( dc, mdc, x, y, m_PadsBottomRight );

    // clean up
    mdc.SelectObject( wxNullBitmap );
}

/////////////////////////
// Draw a pad
////////////////////////////////////////////////////
void wzGUIBoardView::PaintPad( wxDC& dc, wxMemoryDC& mdc, 
                               const short& x, const short& y, wxBitmap pads[3] )
{
    wzSTONE stone = m_pBoard->GetStone( x, y );
    if( stone == wzBLACK )
        mdc.SelectObject( pads[BALCK] );
    else if( stone == wzWHITE )
        mdc.SelectObject( pads[WHITE] );
    else
        mdc.SelectObject( pads[EMPTY] );

    dc.Blit( x*m_PadWidth, y*m_PadWidth, m_PadWidth, m_PadWidth, &mdc, 0, 0 );
}

////////////////////////////////////////////////////
void wzGUIBoardView::HiglightWinningRow( wxDC& dc )
{
    wzBound WinningRow = m_pBoard->GetWinningRow();

    // draw a red dashed line, linking winning row
    short  Xstart = WinningRow.xMin*m_PadWidth + m_Half;
    short  Ystart = WinningRow.yMin*m_PadWidth + m_Half;
    short  Xend   = WinningRow.xMax*m_PadWidth + m_Half;
    short  Yend   = WinningRow.yMax*m_PadWidth + m_Half;

    wxPen pen( wxT("RED"), m_WidePen, wxSHORT_DASH );
    dc.SetPen( pen );

    dc.DrawLine( Xstart, Ystart, Xend, Yend );
    dc.SetPen( wxNullPen );
}

////////////////////////////////////////////////////
void wzGUIBoardView::HiglightLastMove( wxDC& dc )
{
    wzMove move = m_pBoard->GetLastMove();

    // draw a yellow circle
    short  xc = move.X*m_PadWidth + m_Half;
    short  yc = move.Y*m_PadWidth + m_Half;

    wxPen   pen  ( wxT("YELLOW"), m_NarrowPen );
    wxBrush brush( *wxTRANSPARENT_BRUSH );
    dc.SetPen  ( pen );
    dc.SetBrush( brush );

#ifdef __WXMSW__
    // the circle looks smaller on wxMSW
    dc.DrawCircle( xc, yc, m_Half + 1 );
#else
    dc.DrawCircle( xc, yc, m_Half );
#endif

    dc.SetPen  ( wxNullPen );
    dc.SetBrush( wxNullBrush );
}

////////////////////////////////////////////////////
void wzGUIBoardView::HiglightHint( wxDC& dc )
{
    wzMove move = m_pBoard->GetHint();

    // draw a yellow cross
    short  xc = move.X*m_PadWidth + m_Half;
    short  yc = move.Y*m_PadWidth + m_Half;
    short  x0 = xc - m_One3rd;
    short  y0 = yc - m_One3rd;
    short  x1 = xc + m_One3rd;
    short  y1 = yc + m_One3rd;

    wxPen  pen( wxT("YELLOW"), m_NarrowPen );
    dc.SetPen( pen );

    dc.DrawLine( x0, y0, x1, y1 );
    dc.DrawLine( x1, y0, x0, y1 );

    dc.SetPen( wxNullPen );
}

#include "res/BlackBottomLeft.xpm"
#include "res/BlackBottom.xpm"
#include "res/BlackBottomRight.xpm"
#include "res/BlackLeft.xpm"
#include "res/Black.xpm"
#include "res/BlackRight.xpm"
#include "res/BlackTopLeft.xpm"
#include "res/BlackTop.xpm"
#include "res/BlackTopRight.xpm"
#include "res/EmptyBottomLeft.xpm"
#include "res/EmptyBottom.xpm"
#include "res/EmptyBottomRight.xpm"
#include "res/EmptyLeft.xpm"
#include "res/Empty.xpm"
#include "res/EmptyRight.xpm"
#include "res/EmptyTopLeft.xpm"
#include "res/EmptyTop.xpm"
#include "res/EmptyTopRight.xpm"
#include "res/WhiteBottomLeft.xpm"
#include "res/WhiteBottom.xpm"
#include "res/WhiteBottomRight.xpm"
#include "res/WhiteLeft.xpm"
#include "res/White.xpm"
#include "res/WhiteRight.xpm"
#include "res/WhiteTopLeft.xpm"
#include "res/WhiteTop.xpm"
#include "res/WhiteTopRight.xpm"

////////////////////////////
// Load board pads.
/////////////////////////////////////////////////////////////
void wzGUIBoardView::LoadBigPads()
{
    m_PadsTopLeft[EMPTY] = wxBitmap( EmptyTopLeft_xpm );
    m_PadsTopLeft[BALCK] = wxBitmap( BlackTopLeft_xpm );
    m_PadsTopLeft[WHITE] = wxBitmap( WhiteTopLeft_xpm );

    m_PadsTopMiddle[EMPTY] = wxBitmap( EmptyTop_xpm );
    m_PadsTopMiddle[BALCK] = wxBitmap( BlackTop_xpm );
    m_PadsTopMiddle[WHITE] = wxBitmap( WhiteTop_xpm );

    m_PadsTopRight[EMPTY] = wxBitmap( EmptyTopRight_xpm );
    m_PadsTopRight[BALCK] = wxBitmap( BlackTopRight_xpm );
    m_PadsTopRight[WHITE] = wxBitmap( WhiteTopRight_xpm );

    m_PadsCentralLeft[EMPTY] = wxBitmap( EmptyLeft_xpm );
    m_PadsCentralLeft[BALCK] = wxBitmap( BlackLeft_xpm );
    m_PadsCentralLeft[WHITE] = wxBitmap( WhiteLeft_xpm );

    m_PadsCentralMiddle[EMPTY] = wxBitmap( Empty_xpm );
    m_PadsCentralMiddle[BALCK] = wxBitmap( Black_xpm );
    m_PadsCentralMiddle[WHITE] = wxBitmap( White_xpm );

    m_PadsCentralRight[EMPTY] = wxBitmap( EmptyRight_xpm );
    m_PadsCentralRight[BALCK] = wxBitmap( BlackRight_xpm );
    m_PadsCentralRight[WHITE] = wxBitmap( WhiteRight_xpm );

    m_PadsBottomLeft[EMPTY] = wxBitmap( EmptyBottomLeft_xpm );
    m_PadsBottomLeft[BALCK] = wxBitmap( BlackBottomLeft_xpm );
    m_PadsBottomLeft[WHITE] = wxBitmap( WhiteBottomLeft_xpm );

    m_PadsBottomMiddle[EMPTY] = wxBitmap( EmptyBottom_xpm );
    m_PadsBottomMiddle[BALCK] = wxBitmap( BlackBottom_xpm );
    m_PadsBottomMiddle[WHITE] = wxBitmap( WhiteBottom_xpm );

    m_PadsBottomRight[EMPTY] = wxBitmap( EmptyBottomRight_xpm );
    m_PadsBottomRight[BALCK] = wxBitmap( BlackBottomRight_xpm );
    m_PadsBottomRight[WHITE] = wxBitmap( WhiteBottomRight_xpm );
}

#include "res/BlackBottomLeftSmall.xpm"
#include "res/BlackBottomSmall.xpm"
#include "res/BlackBottomRightSmall.xpm"
#include "res/BlackLeftSmall.xpm"
#include "res/BlackSmall.xpm"
#include "res/BlackRightSmall.xpm"
#include "res/BlackTopLeftSmall.xpm"
#include "res/BlackTopSmall.xpm"
#include "res/BlackTopRightSmall.xpm"
#include "res/EmptyBottomLeftSmall.xpm"
#include "res/EmptyBottomSmall.xpm"
#include "res/EmptyBottomRightSmall.xpm"
#include "res/EmptyLeftSmall.xpm"
#include "res/EmptySmall.xpm"
#include "res/EmptyRightSmall.xpm"
#include "res/EmptyTopLeftSmall.xpm"
#include "res/EmptyTopSmall.xpm"
#include "res/EmptyTopRightSmall.xpm"
#include "res/WhiteBottomLeftSmall.xpm"
#include "res/WhiteBottomSmall.xpm"
#include "res/WhiteBottomRightSmall.xpm"
#include "res/WhiteLeftSmall.xpm"
#include "res/WhiteSmall.xpm"
#include "res/WhiteRightSmall.xpm"
#include "res/WhiteTopLeftSmall.xpm"
#include "res/WhiteTopSmall.xpm"
#include "res/WhiteTopRightSmall.xpm"

////////////////////////////
// Load board pads.
/////////////////////////////////////////////////////////////
void wzGUIBoardView::LoadSmallPads()
{
    m_PadsTopLeft[EMPTY] = wxBitmap( EmptyTopLeftSmall_xpm );
    m_PadsTopLeft[BALCK] = wxBitmap( BlackTopLeftSmall_xpm );
    m_PadsTopLeft[WHITE] = wxBitmap( WhiteTopLeftSmall_xpm );

    m_PadsTopMiddle[EMPTY] = wxBitmap( EmptyTopSmall_xpm );
    m_PadsTopMiddle[BALCK] = wxBitmap( BlackTopSmall_xpm );
    m_PadsTopMiddle[WHITE] = wxBitmap( WhiteTopSmall_xpm );

    m_PadsTopRight[EMPTY] = wxBitmap( EmptyTopRightSmall_xpm );
    m_PadsTopRight[BALCK] = wxBitmap( BlackTopRightSmall_xpm );
    m_PadsTopRight[WHITE] = wxBitmap( WhiteTopRightSmall_xpm );

    m_PadsCentralLeft[EMPTY] = wxBitmap( EmptyLeftSmall_xpm );
    m_PadsCentralLeft[BALCK] = wxBitmap( BlackLeftSmall_xpm );
    m_PadsCentralLeft[WHITE] = wxBitmap( WhiteLeftSmall_xpm );

    m_PadsCentralMiddle[EMPTY] = wxBitmap( EmptySmall_xpm );
    m_PadsCentralMiddle[BALCK] = wxBitmap( BlackSmall_xpm );
    m_PadsCentralMiddle[WHITE] = wxBitmap( WhiteSmall_xpm );

    m_PadsCentralRight[EMPTY] = wxBitmap( EmptyRightSmall_xpm );
    m_PadsCentralRight[BALCK] = wxBitmap( BlackRightSmall_xpm );
    m_PadsCentralRight[WHITE] = wxBitmap( WhiteRightSmall_xpm );

    m_PadsBottomLeft[EMPTY] = wxBitmap( EmptyBottomLeftSmall_xpm );
    m_PadsBottomLeft[BALCK] = wxBitmap( BlackBottomLeftSmall_xpm );
    m_PadsBottomLeft[WHITE] = wxBitmap( WhiteBottomLeftSmall_xpm );

    m_PadsBottomMiddle[EMPTY] = wxBitmap( EmptyBottomSmall_xpm );
    m_PadsBottomMiddle[BALCK] = wxBitmap( BlackBottomSmall_xpm );
    m_PadsBottomMiddle[WHITE] = wxBitmap( WhiteBottomSmall_xpm );

    m_PadsBottomRight[EMPTY] = wxBitmap( EmptyBottomRightSmall_xpm );
    m_PadsBottomRight[BALCK] = wxBitmap( BlackBottomRightSmall_xpm );
    m_PadsBottomRight[WHITE] = wxBitmap( WhiteBottomRightSmall_xpm );
}
