// wzGUIFrame.cpp - Frame window for wx GUI app
// Copyright (C) 2006-2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif

#include <wx/filename.h>

// ToggleView doesn't work consistently, because 
// SetClientSize() behaves oddly on wxGTK
#ifdef __WXMSW__
#define __USE_TOGGLE_VIEW__
#endif

#include "wzGUIFrame.h"
#include "wzIDs.h"
#include "wzGUIStrings.h"
#include "wzGUIGame.h"
#include "wzGUIBoardView.h"
#include "wzEvent.h"
#include "wzOptionsDlg.h"
#include "wzAboutDlg.h"
#include "wzLuaOutputDlg.h"
#include "wzAppConfig.h"

#include "res/wzLogo16.xpm"
#include "res/wzLogo32.xpm"
#include "res/NewGame.xpm"
#include "res/OpenGame.xpm"
#include "res/SaveGame.xpm"
#include "res/SaveGameAs.xpm"
#include "res/StopGame.xpm"
#include "res/UndoMove.xpm"
#include "res/HintMove.xpm"
#include "res/FirstPly.xpm"
#include "res/PrevPly.xpm"
#include "res/NextPly.xpm"
#include "res/LastPly.xpm"
#include "res/Options.xpm"
#ifdef __USE_TOGGLE_VIEW__
#include "res/ToggleView.xpm"
#endif
#include "res/ToggleSound.xpm"
#include "res/About.xpm"


BEGIN_EVENT_TABLE( wzGUIFrame, wxFrame )
    EVT_MENU( wzID_NEW_GAME,     wzGUIFrame::OnNewGame )
    EVT_MENU( wzID_OPEN_GAME,    wzGUIFrame::OnOpenGame )
    EVT_MENU( wzID_SAVE_GAME,    wzGUIFrame::OnSaveGame )
    EVT_MENU( wzID_SAVE_GAME_AS, wzGUIFrame::OnSaveGameAs )
    EVT_MENU( wzID_STOP_GAME,    wzGUIFrame::OnStopGame )
    EVT_MENU( wzID_UNDO_MOVE,    wzGUIFrame::OnUndoMove )
    EVT_MENU( wzID_HINT_MOVE,    wzGUIFrame::OnHintMove )
    EVT_MENU( wzID_FIRST_PLY,    wzGUIFrame::OnFirstPly )
    EVT_MENU( wzID_PREV_PLY,     wzGUIFrame::OnPrevPly )
    EVT_MENU( wzID_NEXT_PLY,     wzGUIFrame::OnNextPly )
    EVT_MENU( wzID_LAST_PLY,     wzGUIFrame::OnLastPly )
    EVT_MENU( wzID_OPTIONS,      wzGUIFrame::OnOptions )
    EVT_MENU( wzID_TOGGLE_VIEW,  wzGUIFrame::OnToggleView )
    EVT_MENU( wzID_TOGGLE_SOUND, wzGUIFrame::OnToggleSound )
    EVT_MENU( wxID_ABOUT,        wzGUIFrame::OnAbout )
    EVT_TIMER( wzID_STATUS_TIMER,wzGUIFrame::OnTimer )
    EVT_TIMER( wzID_CLOCK_TICKER,wzGUIFrame::OnTimer )
    EVT_CLOSE(          wzGUIFrame::OnClose )
    EVT_WZ_GAME_IDLE(   wzGUIFrame::OnGameIdle )
    EVT_WZ_GAME_RUN(    wzGUIFrame::OnGameRun )
    EVT_WZ_GAME_END(    wzGUIFrame::OnGameEnd )
    EVT_WZ_GAME_REPLAY( wzGUIFrame::OnGameReplay )
    EVT_WZ_LUA_OUTPUT ( wzGUIFrame::OnLuaOutput  )
    EVT_WZ_FILE_CHANGE( wzGUIFrame::OnFileChange )
    EVT_WZ_FILE_ERROR(  wzGUIFrame::OnFileError )
END_EVENT_TABLE()


/////////////////////////////////////////////
wzGUIFrame::wzGUIFrame()
  : wxFrame( NULL, wxID_ANY, wzAppName, wxDefaultPosition, wxDefaultSize, 
             wxDEFAULT_FRAME_STYLE & ~(wxRESIZE_BORDER|wxRESIZE_BOX|wxMAXIMIZE_BOX) ),
    m_pBoardView   ( new wzGUIBoardView( this ) ),
    m_pGame        ( new wzGUIGame( this, m_pBoardView ) ),
    m_pToolBar     ( NULL ),
    m_pLuaOutputDlg( new wzLuaOutputDlg() ),
    m_StatusTimer  ( this, wzID_STATUS_TIMER ),
    m_ClockTicker  ( this, wzID_CLOCK_TICKER ),
    m_Exit( false )
{
    wxIconBundle Icons;
    Icons.AddIcon( wxIcon(wzLogo16_xpm) );
    Icons.AddIcon( wxIcon(wzLogo32_xpm) );
    SetIcons( Icons );

    CreateToolbar();
    IdleToolbar();

    CreateStatusbar();
    IdleStatusbar();
    m_StatusTimer.Start( 3000 );
    UpdateClocks();

    LoadConfigFile();
}

/////////////////////////////////
wzGUIFrame::~wzGUIFrame()
{
    delete m_pGame;
    m_pLuaOutputDlg->Destroy(); // orphan dialog
}

///////////////////////////////////////////////
void wzGUIFrame::CreateToolbar()
{
    m_pToolBar = CreateToolBar( wxTB_FLAT, wxID_ANY );
    m_pToolBar->CentreOnParent( wxHORIZONTAL );

    m_pToolBar->AddSeparator();
    m_pToolBar->AddTool( wzID_NEW_GAME,  wxBitmap(NewGame_xpm),  _("New"),  _("Start new game") );
    m_pToolBar->AddTool( wzID_OPEN_GAME, wxBitmap(OpenGame_xpm), _("Open"), _("Open saved game") );
    m_pToolBar->AddTool( wzID_SAVE_GAME, wxBitmap(SaveGame_xpm), _("Save"), _("Save game") );
    m_pToolBar->AddTool( wzID_SAVE_GAME_AS, wxBitmap(SaveGameAs_xpm), _("Save As"), _("Save game as") );

    m_pToolBar->AddSeparator();
    m_pToolBar->AddTool( wzID_STOP_GAME, wxBitmap(StopGame_xpm), _("Stop"), _("Stop game") );
    m_pToolBar->AddTool( wzID_UNDO_MOVE, wxBitmap(UndoMove_xpm), _("Undo"), _("Undo move") );
    m_pToolBar->AddTool( wzID_HINT_MOVE, wxBitmap(HintMove_xpm), _("Hint"), _("Hint move") );

    m_pToolBar->AddSeparator();
    m_pToolBar->AddTool( wzID_FIRST_PLY, wxBitmap(FirstPly_xpm), _("First"),    _("Go to first ply") );
    m_pToolBar->AddTool( wzID_PREV_PLY,  wxBitmap(PrevPly_xpm),  _("Previous"), _("Step back") );
    m_pToolBar->AddTool( wzID_NEXT_PLY,  wxBitmap(NextPly_xpm),  _("Next"),     _("Step forward") );
    m_pToolBar->AddTool( wzID_LAST_PLY,  wxBitmap(LastPly_xpm),  _("Last"),     _("Go to last ply") );

    m_pToolBar->AddSeparator();
    m_pToolBar->AddTool( wzID_OPTIONS, wxBitmap(Options_xpm), _("Options"), _("Change options") );

    m_pToolBar->AddSeparator();
#ifdef __USE_TOGGLE_VIEW__
    m_pToolBar->AddCheckTool( wzID_TOGGLE_VIEW,  wxEmptyString, wxBitmap(ToggleView_xpm), wxNullBitmap,
                              _("Window Size"), _("Toggle window size") );
#endif
    m_pToolBar->AddCheckTool( wzID_TOGGLE_SOUND, wxEmptyString, wxBitmap(ToggleSound_xpm), wxNullBitmap, 
                              _("Sound"), _("Toggle sound") );

    m_pToolBar->AddSeparator();
    m_pToolBar->AddTool( wxID_ABOUT, wxBitmap(About_xpm), _("About"), _("About ")+wzAppName );
    m_pToolBar->AddSeparator();

    m_pToolBar->Realize();
}

//////////////////////////////////////
void wzGUIFrame::CreateStatusbar()
{
    // four variable width fields
    int widths[4];
    widths[0] = -1;  // 1/6 of the bar
    widths[1] = -3;  // 2/3
    widths[2] = -1;  // 1/6 
    widths[3] = -1;  // 1/6 
    wxStatusBar* pStatusBar = CreateStatusBar( 4 );
    pStatusBar->SetStatusWidths( 4, widths );
}

/////////////////////////////
void wzGUIFrame::LoadConfigFile()
{
    // frame window position
    wxPoint  position;
    bool     bBigview;
    wzAppConfig AppConfig;
    AppConfig.ReadFrameConfig( position, bBigview );

    Move( position );

    // frame window size
    int width = m_pBoardView->Width();
    if( ( bBigview && !m_pBoardView->IsBigView()) || 
        (!bBigview &&  m_pBoardView->IsBigView())  ) 
        width = m_pBoardView->ToggleView();

    SetClientSize( width, width );
#ifdef __USE_TOGGLE_VIEW__
    m_pToolBar->ToggleTool( wzID_TOGGLE_VIEW, bBigview );
#endif

    // game options
    wzOptions options;
    AppConfig.ReadGameOptions( options );
    m_pGame->SetOptions( options );

    // LuaOutputDlg position and size
    wxSize size;
    AppConfig.ReadLuaOutputConfig( position, size );
    m_pLuaOutputDlg->SetSize( position.x, position.y, size.GetWidth(), size.GetHeight() );

    // tick sound on/off
    bool bSound;
    AppConfig.ReadTickSoundConfig( bSound );
    m_pGame->EnableSound( bSound );
    m_pToolBar->ToggleTool( wzID_TOGGLE_SOUND, bSound );
}

/////////////////////////////
void wzGUIFrame::SaveConfigFile()
{
    // frame window position and size
    wzAppConfig  AppConfig;
    wxPoint  position = GetPosition();
    AppConfig.WriteFrameConfig( position, m_pBoardView->IsBigView() );

    // game options
    wzOptions  options = m_pGame->GetOptions();
    AppConfig.WriteGameOptions( options );

    // LuaOutputDlg position and size
    position = m_pLuaOutputDlg->GetPosition();
    wxSize size = m_pLuaOutputDlg->GetSize();
    AppConfig.WriteLuaOutputConfig( position, size );

    // tick sound on/off
    AppConfig.WriteTickSoundConfig( m_pToolBar->GetToolState( wzID_TOGGLE_SOUND ) );
}

//////////////////////////////////////
void wzGUIFrame::OnNewGame( wxCommandEvent& event )
{
    m_pLuaOutputDlg->Clean();
    m_pGame->NewGame();
}

//////////////////////////////////////
void wzGUIFrame::OnOpenGame( wxCommandEvent& event )
{
    m_pLuaOutputDlg->Clean();
    m_pGame->OpenGame();
}

//////////////////////////////////////
void wzGUIFrame::OnSaveGame( wxCommandEvent& event )
{
    m_pGame->SaveGame();
}

//////////////////////////////////////
void wzGUIFrame::OnSaveGameAs( wxCommandEvent& event )
{
    m_pGame->SaveGameAs();
}

//////////////////////////////////////
void wzGUIFrame::OnStopGame( wxCommandEvent& event )
{
    m_pGame->StopGame();
}

//////////////////////////////////////
void wzGUIFrame::OnUndoMove( wxCommandEvent& event )
{
    m_pGame->UndoMove();
}

//////////////////////////////////////
void wzGUIFrame::OnHintMove( wxCommandEvent& event )
{
    m_pGame->HintMove();
}

//////////////////////////////////////
void wzGUIFrame::OnFirstPly( wxCommandEvent& event )
{
    m_pGame->FirstPly();
}

//////////////////////////////////////
void wzGUIFrame::OnPrevPly( wxCommandEvent& event )
{
    m_pGame->PrevPly();
}

//////////////////////////////////////
void wzGUIFrame::OnNextPly( wxCommandEvent& event )
{
    m_pGame->NextPly();
}

//////////////////////////////////////
void wzGUIFrame::OnLastPly( wxCommandEvent& event )
{
    m_pGame->LastPly();
}

//////////////////////////////////////
void wzGUIFrame::OnOptions( wxCommandEvent& event )
{
    wzOptions options = m_pGame->GetOptions();
    wzOptionsDlg dlg( this, options );
    while( dlg.ShowModal() == wxID_OK )
    {
        dlg.GetOptions( options );
        if( m_pGame->SetOptions( options ) )
        {
            break;
        }
        else
        {
            wxString msg = _("Some field(s) is empty.");
            wxMessageBox( msg, wzAppName, wxOK|wxICON_ERROR );
        }
    }
}

//////////////////////////////////////
void wzGUIFrame::OnToggleView( wxCommandEvent& event )
{
    int size = m_pBoardView->ToggleView();
    SetClientSize( size, size );
}

//////////////////////////////////////
void wzGUIFrame::OnToggleSound( wxCommandEvent& event )
{
    m_pGame->EnableSound( m_pToolBar->GetToolState( wzID_TOGGLE_SOUND ) );
}

//////////////////////////////////////
void wzGUIFrame::OnAbout( wxCommandEvent& event )
{
    wzAboutDlg dlg( this );
    dlg.ShowModal();
}

//////////////////////////////////////
void wzGUIFrame::OnClose( wxCloseEvent& event )
{
    // game still running
    if( m_pGame->IsRunning() ) 
    {
        // refuse to close
        if( event.CanVeto() )  event.Veto();

        // stop game and set exiting flag
        m_pGame->StopGame();
        m_Exit = true;

    }
    else // game ended
    {
        SaveConfigFile();

        // actually close the window
        Destroy();
    }
}

//////////////////////////
void wzGUIFrame::OnGameIdle( wzGameIdleEvent& event )
{
    UpdateClocks();
    IdleStatusbar();
    IdleToolbar();
}

//////////////////////////
void wzGUIFrame::OnGameRun( wzGameRunEvent& event )
{
    RunningStatusbar();
    RunningToolbar();
    m_ClockTicker.Start( 500 );
}

//////////////////////////
// Game end with one of situations:
//   - Black wins
//   - White wins
//   - Game draw
//   - Game stopped
//////////////////////////////////////////////////////
void wzGUIFrame::OnGameEnd( wzGameEndEvent& event )
{
    // flag already set
    if( m_Exit )
    {
        // force the window to be closed by sending
        // wxCloseEvent with vetoing disabled. 
        Close( true );
        return;
    }

    EndedStatusbar( event.GetId() );
    EndedToolbar();

    m_ClockTicker.Stop();
    UpdateClocks();
}

//////////////////////////
void wzGUIFrame::OnGameReplay( wzGameReplayEvent& event )
{
    ReplayingStatusbar();
    ReplayingToolbar();
}

///////////////////////
// Append msg to LuaOutputDlg
////////////////////////////////////////////////////////
void wzGUIFrame::OnLuaOutput( wzLuaOutputEvent& event )
{
    m_pLuaOutputDlg->Append( event.GetMessage() );
}

///////////////////////
// update title bar
/////////////////////////////////////////////////
void wzGUIFrame::OnFileChange( wzFileChangeEvent& event )
{
    wxString FileName( "untitled.tick5" );

    wxString Path = (m_pGame->GetFileName()).c_str();
    if( !Path.IsEmpty() )
        FileName = StripPath( Path );

    wxString Title = wzAppName + wxT( " - " ) + FileName;
    if( !m_pGame->IsSaved() )
        Title += wxT( " *" );

    SetTitle( Title );
}

////////////////////////////////////////////////////////
void wzGUIFrame::OnFileError( wzFileErrorEvent& event )
{
    wxMessageBox( event.GetMessage(), wzAppName, wxOK|wxICON_ERROR );
}

/////////////////////////////////////////////////
void wzGUIFrame::OnTimer( wxTimerEvent& event )
{
    if( event.GetId() == wzID_CLOCK_TICKER )
        UpdateClocks();
    else if( event.GetId() == wzID_STATUS_TIMER )
        FlipStatusMsg();
}

//////////////////////////////////////
void wzGUIFrame::FlipStatusMsg()
{
    static int iMsg = 0;
    SetStatusText( m_StatusMsg[iMsg], MSG_FIELD );

    if( ++iMsg > 1 )  iMsg = 0;
}

//////////////////////////////////////
void wzGUIFrame::UpdateClocks()
{
    std::string stime = m_pGame->GetPlayTime( 0 );
    wxString msg = stime.c_str();
    SetStatusText( msg, CLOCK_FIELD0 );

    stime = m_pGame->GetPlayTime( 1 );
    msg = stime.c_str();
    SetStatusText( msg, CLOCK_FIELD1 );}

//////////////////////////////////////
void wzGUIFrame::IdleStatusbar()
{
    m_StatusMsg[0] = _( "Welcom to " ) + wzAppName;
    m_StatusMsg[1] = _( "Start a new game or open a saved one?" );

    SetStatusText( m_StatusMsg[0], MSG_FIELD );
}

//////////////////////////
// Set toolbar to idle state
////////////////////////////////////////////////////
void wzGUIFrame::IdleToolbar()
{
    m_pToolBar->EnableTool( wzID_NEW_GAME,  true );
    m_pToolBar->EnableTool( wzID_OPEN_GAME, true );
    m_pToolBar->EnableTool( wzID_SAVE_GAME, false );
    m_pToolBar->EnableTool( wzID_SAVE_GAME_AS, false );
    m_pToolBar->EnableTool( wzID_STOP_GAME, false );
    m_pToolBar->EnableTool( wzID_UNDO_MOVE, false );
    m_pToolBar->EnableTool( wzID_HINT_MOVE, false );
    m_pToolBar->EnableTool( wzID_FIRST_PLY, false );
    m_pToolBar->EnableTool( wzID_PREV_PLY,  false );
    m_pToolBar->EnableTool( wzID_NEXT_PLY,  false );
    m_pToolBar->EnableTool( wzID_LAST_PLY,  false );
    m_pToolBar->EnableTool( wzID_OPTIONS,   true);
}

//////////////////////////
// set status bar to running state
/////////////////////////////////////////////////////////
void wzGUIFrame::RunningStatusbar()
{
    m_StatusMsg[0] = _( "Game is running ..." );
    SetStatusText( m_StatusMsg[0], MSG_FIELD );

    wzOptions options = m_pGame->GetOptions();

    wxString BlackName = options.m_BlackName.c_str();
    BlackName = StripPath( BlackName );

    wxString WhiteName = options.m_WhiteName.c_str();
    WhiteName = StripPath( WhiteName );

    m_StatusMsg[1] = BlackName + _("   vs.   ") + WhiteName; 
}

////////////////////////
// set toolbar to running state
//////////////////////////////////////////////////////
void wzGUIFrame::RunningToolbar()
{
    m_pToolBar->EnableTool( wzID_NEW_GAME,  false );
    m_pToolBar->EnableTool( wzID_OPEN_GAME, false );
    m_pToolBar->EnableTool( wzID_SAVE_GAME, true );
    m_pToolBar->EnableTool( wzID_SAVE_GAME_AS, true );
    m_pToolBar->EnableTool( wzID_STOP_GAME, true );
    m_pToolBar->EnableTool( wzID_FIRST_PLY, false );
    m_pToolBar->EnableTool( wzID_PREV_PLY,  false );
    m_pToolBar->EnableTool( wzID_NEXT_PLY,  false );
    m_pToolBar->EnableTool( wzID_LAST_PLY,  false );
    m_pToolBar->EnableTool( wzID_OPTIONS,   false);

    wzOptions options = m_pGame->GetOptions();
    bool bHumanPlayer = options.HasHumanPlayer();
    m_pToolBar->EnableTool( wzID_UNDO_MOVE, bHumanPlayer );
    m_pToolBar->EnableTool( wzID_HINT_MOVE, bHumanPlayer );
}

/////////////////////////////
// set status bar to ended state
/////////////////////////////////////////////////////////////////
void wzGUIFrame::EndedStatusbar( const int ID )
{
    switch( ID )
    {
    case wzGAME_BLACK_WIN :
    case wzGAME_WHITE_WIN :
      {
        wzOptions options = m_pGame->GetOptions();
        wxString WinnerName = (ID == wzGAME_BLACK_WIN)? options.m_BlackName.c_str() : 
                                                        options.m_WhiteName.c_str();
        m_StatusMsg[0] = StripPath( WinnerName ) + _( " wins." );
        SetStatusText( m_StatusMsg[0], MSG_FIELD );

        break;
      }

    case wzGAME_STOPPED :
    case wzGAME_DRAW :
        m_StatusMsg[0] = (ID == wzGAME_STOPPED)? _( "Game is stopped." ) : 
                                                 _( "Game is a draw." );
        SetStatusText( m_StatusMsg[0], MSG_FIELD );
        wxMessageBox( m_StatusMsg[0], wzAppName, wxOK|wxICON_INFORMATION );
    }
}

/////////////////////////////
// set toolbar to ended state
//////////////////////////////////////////////////
void wzGUIFrame::EndedToolbar()
{
    m_pToolBar->EnableTool( wzID_NEW_GAME,  true );
    m_pToolBar->EnableTool( wzID_OPEN_GAME, true );
    m_pToolBar->EnableTool( wzID_SAVE_GAME, true );
    m_pToolBar->EnableTool( wzID_SAVE_GAME_AS, true );
    m_pToolBar->EnableTool( wzID_STOP_GAME, false );
    m_pToolBar->EnableTool( wzID_UNDO_MOVE, false );
    m_pToolBar->EnableTool( wzID_HINT_MOVE, false );
    m_pToolBar->EnableTool( wzID_FIRST_PLY, true );
    m_pToolBar->EnableTool( wzID_PREV_PLY,  false );
    m_pToolBar->EnableTool( wzID_NEXT_PLY,  false );
    m_pToolBar->EnableTool( wzID_LAST_PLY,  false );
    m_pToolBar->EnableTool( wzID_OPTIONS,   true);
}

//////////////////////////
// set status bar to replaying state
////////////////////////////////////////////////////////
void wzGUIFrame::ReplayingStatusbar()
{
    m_StatusMsg[0] = _("Game is replaying ...");
    SetStatusText( m_StatusMsg[0], MSG_FIELD );
}

//////////////////////////
// set toolbar to replaying state
////////////////////////////////////////////////////////
void wzGUIFrame::ReplayingToolbar()
{
    m_pToolBar->EnableTool( wzID_NEW_GAME,  true );
    m_pToolBar->EnableTool( wzID_OPEN_GAME, true );
    m_pToolBar->EnableTool( wzID_SAVE_GAME, false );
    m_pToolBar->EnableTool( wzID_SAVE_GAME_AS, true );
    m_pToolBar->EnableTool( wzID_STOP_GAME, false );
    m_pToolBar->EnableTool( wzID_UNDO_MOVE, false );
    m_pToolBar->EnableTool( wzID_HINT_MOVE, false );
    m_pToolBar->EnableTool( wzID_FIRST_PLY, true );
    m_pToolBar->EnableTool( wzID_PREV_PLY,  true );
    m_pToolBar->EnableTool( wzID_NEXT_PLY,  true );
    m_pToolBar->EnableTool( wzID_LAST_PLY,  true );
    m_pToolBar->EnableTool( wzID_OPTIONS,   true);
}

///////////////////////////////////////////////////
wxString wzGUIFrame::StripPath( wxString& filename )
{
    wxFileName path( filename );
    return path.GetFullName();
}
