// wzClockTest.cpp - Unit test for wzClock
// Copyright (C) 2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <TestSuite.h>
#include <TestCaller.h>

#include "wzClockTest.h"
#include <wzClock.h>


void wzClockTest::testCtor()
{
    wzClock clock;
    int  hours, minutes, seconds, decisecs;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 0 == decisecs );
}

void wzClockTest::testReset()
{
    wzClock clock;
    int  hours, minutes, seconds, decisecs;

    clock.Start();
    Delay( 110 );
    clock.Stop();

    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 1 == decisecs );

    clock.Reset();
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 0 == decisecs );
}

void wzClockTest::testConvert()
{
    wzClock clock;
    int  hours, minutes, seconds, decisecs;

    // 1) pretend 0.9 sec elapsed
    clock.m_Elapsed = 90;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 9 == decisecs );

    // pretend 0.2 sec passed
    clock.m_Elapsed += 20;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 1 == seconds );
    assert( 1 == decisecs );


    // 2) pretend 59.9 sec elapsed
    clock.m_Elapsed = 5990;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert(  0 == hours );
    assert(  0 == minutes );
    assert( 59 == seconds );
    assert(  9 == decisecs );

    // pretend 0.2 sec passed
    clock.m_Elapsed += 20;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 1 == minutes );
    assert( 0 == seconds );
    assert( 1 == decisecs );

    // 3) pretend 59:59.9 elapsed
    clock.m_Elapsed = 59*6000 + 5990;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert(  0 == hours );
    assert( 59 == minutes );
    assert( 59 == seconds );
    assert(  9 == decisecs );

    // pretend 0.2 sec passed
    clock.m_Elapsed += 20;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 1 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 1 == decisecs );

    // 4) pretend 0.06 sec less than limit
    clock.m_Elapsed = wzClock::m_Limit - 6;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( MAX_HOURS    == hours );
    assert( MAX_MINUTES  == minutes );
    assert( MAX_SECONDS  == seconds );
    assert( MAX_DECISECS == decisecs + 1 );

    // 5) pretend 0.05 sec less than limit
    clock.m_Elapsed = wzClock::m_Limit - 5;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( MAX_HOURS    == hours );
    assert( MAX_MINUTES  == minutes );
    assert( MAX_SECONDS  == seconds );
    assert( MAX_DECISECS == decisecs );  // round up

    // 6) pretend limit reached
    clock.m_Elapsed = wzClock::m_Limit;
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( MAX_HOURS    == hours );
    assert( MAX_MINUTES  == minutes );
    assert( MAX_SECONDS  == seconds );
    assert( MAX_DECISECS == decisecs );
}

void wzClockTest::testStart()
{
    wzClock clock;
    int  hours, minutes, seconds, decisecs;

    clock.Start();
    Delay( 130 );
    clock.Start(); // effectless and harmless
    clock.Stop();
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 1 == decisecs );

    // 0.01 second less than limit
    clock.m_Elapsed = wzClock::m_Limit - 1;
    assert( !clock.ExceedLimit() );

    clock.Start();
    Delay( 100 );
    clock.Stop();
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( MAX_HOURS    == hours );
    assert( MAX_MINUTES  == minutes );
    assert( MAX_SECONDS  == seconds );
    assert( MAX_DECISECS == decisecs );

    // reached limit
    assert( clock.ExceedLimit() );

    clock.Start(); // now, effectless and harmless
    Delay( 50 );
    clock.Stop();
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( MAX_HOURS    == hours );
    assert( MAX_MINUTES  == minutes );
    assert( MAX_SECONDS  == seconds );
    assert( MAX_DECISECS == decisecs );
}

void wzClockTest::testStop()
{
    wzClock clock;
    int  hours, minutes, seconds, decisecs;

    clock.Start();
    Delay( 160 ); // 0.16 sec
    clock.Stop();
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 2 == decisecs );

    clock.Start();
    Delay( 1000 ); // 0.16 + 1 = 1.16 sec
    clock.Stop();
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 1 == seconds );
    assert( 2 == decisecs );

    clock.Start();
    Delay( 200 ); // 1.16 + 0.2 = 1.36 sec
    clock.Stop();
    clock.Convert( hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 1 == seconds );
    assert( 4 == decisecs );
}
    
void wzClockTest::testGetTime()
{
    wzClock clock;
    std::string stime;
    int  hours, minutes, seconds, decisecs;

    clock.Start();
    Delay( 150 );  // 0.15 sec
    stime = clock.GetTime();
    Decode( stime, hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 0 == seconds );
    assert( 2 == decisecs );  // round up

    Delay( 1000 ); // 0.15 + 1 = 1.15 sec
    stime = clock.GetTime();
    Decode( stime, hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 1 == seconds );
    assert( 2 == decisecs );

    Delay( 200 ); // 1.15 + 0.2 = 1.35 sec
    stime = clock.GetTime();
    Decode( stime, hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 1 == seconds );
    assert( 4 == decisecs );

    Delay( 260 ); // 1.35 + 0.26 = 1.61 sec
    clock.Stop();
    stime = clock.GetTime();
    Decode( stime, hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 1 == seconds );
    assert( 6 == decisecs );  // round down

    // already stopped
    Delay( 50 );
    stime = clock.GetTime();
    Decode( stime, hours, minutes, seconds, decisecs );
    assert( 0 == hours );
    assert( 0 == minutes );
    assert( 1 == seconds );
    assert( 6 == decisecs );  // round down
}

// also help checking gettimeofday()
// cannot be a static member to call assert() defined in CppUnit
//////////////////////////////////////////////////////
void wzClockTest::Delay( long millisecs )
{
    timeval  start, stop;
    assert( 0 == gettimeofday( &start, NULL ) );

    long millidiff = 0L;
    while( millidiff < millisecs )
    {
        assert( 0 == gettimeofday( &stop, NULL ) );

        millidiff =  (stop.tv_sec  - start.tv_sec)*1000;
        millidiff += (stop.tv_usec - start.tv_usec)/1000;
    }
}

void wzClockTest::Decode( std:: string stime, int& hours, int& minutes, 
                          int& seconds, int& decisecs )
{
    hours    = atoi( stime.substr(0,1).c_str() );
    minutes  = atoi( stime.substr(3,4).c_str() );
    seconds  = atoi( stime.substr(6,7).c_str() );
    decisecs = atoi( stime.substr(9,10).c_str());
}

Test* wzClockTest::suite()
{
    TestSuite *suite = new TestSuite( "wzClockTest" );
    suite->addTest( CU_SUIT_TEST( wzClockTest, testCtor ) );
    suite->addTest( CU_SUIT_TEST( wzClockTest, testReset ) );
    suite->addTest( CU_SUIT_TEST( wzClockTest, testConvert ) );
    suite->addTest( CU_SUIT_TEST( wzClockTest, testStart ) );
    suite->addTest( CU_SUIT_TEST( wzClockTest, testStop ) );
    suite->addTest( CU_SUIT_TEST( wzClockTest, testGetTime ) );

    return suite;
}
