// wzGameFileTest.cpp - Unit test for wzGameFile
// Copyright (C) 2007  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <TestSuite.h>
#include <TestCaller.h>

#include "wzGameFileTest.h"
#include <wzGameFile.h>
#include <wzGame.h>
#include <wzBoard.h>
#include <wzVersion.h>

#include <fstream>
using namespace std;


void wzGameFileTest::setUp()
{
    m_pGame     = new wzGame();
    m_pGameFile = new wzGameFile( m_pGame );
}

void wzGameFileTest::tearDown()
{
    delete m_pGameFile;
    delete m_pGame;
}

void wzGameFileTest::testWrite()
{
    ofstream  ofstr( "untitled.tick5", ios::out|ios::binary|ios::trunc );
    try
    {
        m_pGameFile->Write( ofstr );
    }
    catch( ... )
    {
        // fail: writing error
        assert( false );
    }

    ofstr.close();

    // open file just wrote
    ifstream  ifstr( "untitled.tick5", ios::in|ios::binary );

    // ID
    assert( wzGameFile::m_MagicID[0] == ifstr.get() );
    assert( wzGameFile::m_MagicID[1] == ifstr.get() );

    // version
    wzVersion version;
    assert( version.Major() == static_cast<wzUInt8>( ifstr.get() ) );
    assert( version.Minor() == static_cast<wzUInt8>( ifstr.get() ) );
    assert( version.Micro() == static_cast<wzUInt8>( ifstr.get() ) );

    // options
    wzOptions options;
    ifstr >> options;
    assert( options.Valid() );

    // 8 bytes time: all zeros, as game hasn't started
    for( short i = 0; i < 8; i++ )
    {
        assert( 0 == static_cast<wzUInt8>( ifstr.get() ) );
    }

    // moves: no move made so far
    assert( 0 == static_cast<wzUInt8>( ifstr.get() ) );

    // trailer
    assert( wzGameFile::m_Trailer == ifstr.get() );

    // EOF
    ifstr.get();
    assert( !ifstr );
}

void wzGameFileTest::testWriteMoves()
{
    m_pGameFile->m_pBoard->Clean();
    m_pGameFile->m_pBoard->Set( 7, 7, wzBLACK );  // move #1
    m_pGameFile->m_pBoard->Set( 8, 9, wzWHITE );  // move #2

    ofstream  ofstr( "untitled.tick5", ios::out|ios::binary|ios::trunc );
    try
    {
        m_pGameFile->WriteMoves( ofstr );
    }
    catch( ... )
    {
        // fail: writing error
        assert( false );
    }

    ofstr.close();

    ifstream  ifstr( "untitled.tick5", ios::in|ios::binary );

    // number of moves
    assert( 2 == static_cast<wzUInt8>( ifstr.get() ) );

    // move #1
    assert( 7 == static_cast<wzUInt8>( ifstr.get() ) );
    assert( 7 == static_cast<wzUInt8>( ifstr.get() ) );

    // move #2
    assert( 8 == static_cast<wzUInt8>( ifstr.get() ) );
    assert( 9 == static_cast<wzUInt8>( ifstr.get() ) );
}

void wzGameFileTest::testRead()
{
    ofstream  ofstr( "untitled.tick5", ios::out|ios::binary|ios::trunc );

    // ID, version and options
    ofstr.write( wzGameFile::m_MagicID, 2 );
    ofstr << wzGameFile::m_Version;
    wzOptions options;
    ofstr << options;

    // 8 bytes for time
    wzUInt8  Byte = 9;
    for( short i = 0; i < 8; i++ )  ofstr << Byte;

    // no moves
    wzUInt8  plies = 0;
    ofstr << plies;

    // trailer
    ofstr << wzGameFile::m_Trailer;

    ofstr.close();

    ifstream  ifstr( "untitled.tick5", ios::in|ios::binary );
    try
    {
        m_pGameFile->Read( ifstr );
    }
    catch( ... )
    {
        // fail: reading error
        assert( false );
    }

    // EOF
    try
    {
        ifstr.get();

        // fail: still not reach EOF
        assert( false );
    }
    catch( ios_base::failure )
    {
    }

    ifstr.close();
}

void wzGameFileTest::testReadWrongID()
{
    ofstream ofstr( "untitled.tick5", ios::out|ios::binary|ios::trunc );

    // wrong ID
    ofstr.write( "ZW", 2 );

    ofstr.close();

    ifstream ifstr( "untitled.tick5", ios::in|ios::binary );
    try
    {
        m_pGameFile->Read( ifstr );

        // fail: no exception for wrong ID 
        assert( false );
    }
    catch( const char* msg )
    {
        assert( 0 == strcmp( msg, "Not a valid Tick5 game file." ) );
    }
}

void wzGameFileTest::testReadWrongOptions()
{
    ofstream  ofstr;
    ifstream  ifstr;
    wzOptions options;

    ///////////////////////////////
    // wrong player code
    ///////////////////////////////////////////////
    ofstr.open( "untitled.tick5", ios::out|ios::binary|ios::trunc );

    // ID and version
    ofstr.write( wzGameFile::m_MagicID, 2 );
    ofstr << wzGameFile::m_Version;

    // wrong options: wrong player code
    options.m_BlackStone = (wzCOMPUTER_PLAYER >> 1);
    ofstr << options;

    ofstr.close();

    ifstr.open( "untitled.tick5", ios::in|ios::binary );
    try
    {
        m_pGameFile->Read( ifstr );

        // fail: no exception for wrong options
        assert( false );
    }
    catch( const char* msg )
    {
        assert( 0 == strcmp( msg, "Corrupted file: invalid game options." ) );
    }

    ifstr.close();

    ///////////////////////////////
    // no Lua script name
    ///////////////////////////////////////////////
    ofstr.open( "untitled.tick5", ios::out|ios::binary|ios::trunc );

    // ID and version
    ofstr.write( wzGameFile::m_MagicID, 2 );
    ofstr << wzGameFile::m_Version;

    // wrong options: empty name
    options.m_BlackStone = wzLUA_AI_PLAYER;
    options.m_BlackName.clear();
    ofstr << options;

    // no moves
    wzUInt8 plies = 0;
    ofstr << plies;

    ofstr.close();

    ifstr.open( "untitled.tick5", ios::in|ios::binary );
    try
    {
        m_pGameFile->Read( ifstr );

        // fail: no exception for wrong options
        assert( false );
    }
    catch( const char* msg )
    {
        assert( 0 == strcmp( msg, "Corrupted file: invalid game options." ) );
    }

    ifstr.close();
}

void wzGameFileTest::testApplyMoves()
{
    ofstream ofstr;
    ifstream ifstr;
    ifstr.exceptions( ios::failbit|ios::badbit );

    ////////////////
    // wrong coordinates
    ////////////////////////////////////////
    ofstr.open( "untitled.tick5", ios::out|ios::binary|ios::trunc );

    // 2 moves
    ofstr << static_cast<wzUInt8>( 2 );

    // move #1
    ofstr << static_cast<wzUInt8>( 7 );
    ofstr << static_cast<wzUInt8>( 7 );

    // move #2 : wrong move
    ofstr << static_cast<wzUInt8>( 15 ); // wrong coordinate
    ofstr << static_cast<wzUInt8>( 7 );

    ofstr.close();

    ifstr.open( "untitled.tick5", ios::in|ios::binary );
    try
    {
        m_pGameFile->ApplyMoves( ifstr );

        // fail: no exception for wrong move
        assert( false );
    }
    catch( const char* msg )
    {
        assert( 0 == strcmp( msg, "Corrupted file: invalid move." ) );
    }

    ifstr.close();

    ////////////////
    // same spot 
    ////////////////////////////////////////
    ofstr.open( "untitled.tick5", ios::out|ios::binary|ios::trunc );

    // 2 moves
    ofstr << static_cast<wzUInt8>( 2 );

    // move #1
    ofstr << static_cast<wzUInt8>( 7 );
    ofstr << static_cast<wzUInt8>( 7 );

    // move #2 : wrong move, coordinates same as move #1
    ofstr << static_cast<wzUInt8>( 7 );
    ofstr << static_cast<wzUInt8>( 7 );

    ofstr.close();

    ifstr.open( "untitled.tick5", ios::in|ios::binary );
    try
    {
        m_pGameFile->ApplyMoves( ifstr );

        // fail: no exception for wrong move
        assert( false );
    }
    catch( const char* msg )
    {
        assert( 0 == strcmp( msg, "Corrupted file: invalid move." ) );
    }

    ifstr.close();

    ////////////////
    // number of moves is bigger than moves in the file
    //////////////////////////////////////////////////////
    ofstr.open( "untitled.tick5", ios::out|ios::binary|ios::trunc );

    // number of moves: wrong, should be 2
    ofstr << static_cast<wzUInt8>( 3 );

    // move #1
    ofstr << static_cast<wzUInt8>( 7 );
    ofstr << static_cast<wzUInt8>( 7 );

    // move #2
    ofstr << static_cast<wzUInt8>( 8 );
    ofstr << static_cast<wzUInt8>( 8 );

    ofstr.close();

    ifstr.open( "untitled.tick5", ios::in|ios::binary );
    try
    {
        m_pGameFile->ApplyMoves( ifstr );

        // fail: no exception for wrong number of moves
        assert( false );
    }
    catch( ios_base::failure )
    {
    }

    ifstr.close();
}

//////////////////////
// the saved must be read without error
///////////////////////////////////////////
void wzGameFileTest::testSaveOpen()
{
    m_pGameFile->m_pBoard->Set( 7, 7, wzBLACK );
    m_pGameFile->m_pBoard->Set( 8, 9, wzWHITE ); 

    try
    {
        m_pGameFile->Save();
    }
    catch( ... )
    {
        // fail: writing error
        assert( false );
    }

    m_pGameFile->m_pBoard->Clean();

    try
    {
        m_pGameFile->Open();
    }
    catch( ... )
    {
        // fail: reading error
        assert( false );
    }

    assert( wzBLACK == m_pGameFile->m_pBoard->GetStone( 7, 7 ) );
    assert( wzWHITE == m_pGameFile->m_pBoard->GetStone( 8, 9 ) );
}

//////////////////////
// the saved must be read without error
///////////////////////////////////////////
void wzGameFileTest::testSaveAs()
{
    m_pGameFile->m_pBoard->Set( 7, 7, wzBLACK );
    m_pGameFile->m_pBoard->Set( 8, 9, wzWHITE ); 

    try
    {
        m_pGameFile->SaveAs();
    }
    catch( ... )
    {
        // fail: writing error
        assert( false );
    }

    m_pGameFile->m_pBoard->Clean();

    try
    {
        m_pGameFile->Open();
    }
    catch( ... )
    {
        // fail: reading error
        assert( false );
    }

    assert( wzBLACK == m_pGameFile->m_pBoard->GetStone( 7, 7 ) );
    assert( wzWHITE == m_pGameFile->m_pBoard->GetStone( 8, 9 ) );
}

void wzGameFileTest::testIsSaved()
{
    // initial state
    assert( m_pGameFile->IsSaved() == true );

    // wzBoard::Clean() won't affect the state
    m_pGameFile->m_pBoard->Clean();
    assert( m_pGameFile->IsSaved() == true );

    m_pGameFile->Save();
    assert( m_pGameFile->IsSaved() == true );

    m_pGameFile->m_pBoard->Set( 7, 7, wzBLACK );
    assert( m_pGameFile->IsSaved() == false );

    m_pGameFile->Save();
    assert( m_pGameFile->IsSaved() == true );

    m_pGameFile->m_pBoard->Set( 8, 9, wzWHITE ); 

    m_pGameFile->Save();
    assert( m_pGameFile->IsSaved() == true );

    m_pGameFile->m_pBoard->UndoMove();
    assert( m_pGameFile->IsSaved() == false );

    m_pGameFile->Save();
    assert( m_pGameFile->IsSaved() == true );
}

Test* wzGameFileTest::suite()
{
    TestSuite *suite = new TestSuite( "wzGameFileTest" );

    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testWrite ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testWriteMoves ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testRead ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testReadWrongID ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testReadWrongOptions ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testApplyMoves ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testSaveOpen ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testSaveAs ) );
    suite->addTest( CU_SUIT_TEST( wzGameFileTest, testIsSaved ) );

    return suite;
}

