// wzPatternIteratorTest.h - Unit test for wzPattern
// Copyright (C) 2006  XMartin Yao
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <TestSuite.h>
#include <TestCaller.h>

#include "wzPatternIteratorTest.h"
#include <wzPatternIterator.h>

// traverse a rectanular reange
void wzPatternIteratorTest::testHorizontal5box()
{
    wzHorizontal5 h5;
    wzPatternIterator& it = h5.GetIterator(); 

    // test SetBound(): a point
    wzBound box;
    box.xMin = 7;
    box.yMin = 7;
    box.xMax = 7;
    box.yMax = 7;
    assert( !it.SetBound( box ) );

    // test SetBound(): not wild enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 5;
    box.yMax = 2;
    assert( !it.SetBound( box ) );

    // test SetBound(): not wild enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 3;
    box.yMax = 8;
    assert( !it.SetBound( box ) );

    // test SetBound(): just minimal size
    box.xMin = 5;
    box.yMin = 5;
    box.xMax = 9;
    box.yMax = 5;
    assert( it.SetBound( box ) );
    wzBound row = it.GetRow();
    assert( row.xMin == 5 );
    assert( row.yMin == 5 );
    assert( row.xMax == 9 );
    assert( row.yMax == 5 );

    const wzBitBoard& p = it.GetPattern();
    assert( !p.IsEmpty( 5, 5 ) );
    assert( !p.IsEmpty( 6, 5 ) );
    assert( !p.IsEmpty( 7, 5 ) );
    assert( !p.IsEmpty( 8, 5 ) );
    assert( !p.IsEmpty( 9, 5 ) );

    //
    // test a traverse range: (4,5) - (9,6)
    //
    box.xMin = 4;
    box.yMin = 5;
    box.xMax = 9;
    box.yMax = 6;
    assert( it.SetBound( box ) );

    // 1st element
    row = it.GetRow();
    assert( row.xMin == 4 );
    assert( row.yMin == 5 );
    assert( row.xMax == 8 );
    assert( row.yMax == 5 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 4 );
    assert( row.yMin == 6 );
    assert( row.xMax == 8 );
    assert( row.yMax == 6 );

    // 4th element
    assert( it.Next() );
    const wzBitBoard& p2 = it.GetPattern();
    assert( !p2.IsEmpty( 5, 6 ) );
    assert( !p2.IsEmpty( 6, 6 ) );
    assert( !p2.IsEmpty( 7, 6 ) );
    assert( !p2.IsEmpty( 8, 6 ) );
    assert( !p2.IsEmpty( 9, 6 ) );

    // only 5 bits
    assert( (p2 & p2) == 5 );

    // others are empty, try some of them
    assert( p2.IsEmpty( 5, 5 ) );
    assert( p2.IsEmpty( 6, 5 ) );
    assert( p2.IsEmpty( 7, 5 ) );
    assert( p2.IsEmpty( 8, 5 ) );
    assert( p2.IsEmpty( 9, 5 ) );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    assert( row.xMin == 4 );
    assert( row.yMin == 5 );
    assert( row.xMax == 8 );
    assert( row.yMax == 5 );
}

// traverse a line
void wzPatternIteratorTest::testHorizontal5line()
{
    wzHorizontal5 h5;
    wzPatternIterator& it = h5.GetIterator(); 

    //
    // one element
    //
    assert( it.SetCenter( 0, 3 ) );
    wzBound row = it.GetRow();
    assert( row.xMin == 0 );
    assert( row.yMin == 3 );
    assert( row.xMax == 4 );
    assert( row.yMax == 3 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 14, 10 ) );
    row = it.GetRow();
    assert( row.xMin == 10 );
    assert( row.yMin == 10 );
    assert( row.xMax == 14 );
    assert( row.yMax == 10 );

    // no more
    assert( !it.Next() );

    //
    // 5 elements
    //
    assert( it.SetCenter( 7, 7 ) );

    // 1st element
    row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 7 );
    assert( row.xMax == 7 );
    assert( row.yMax == 7 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );

    // 4th element
    assert( it.Next() );

    // 5th element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 7 );
    assert( row.yMin == 7 );
    assert( row.xMax == 11 );
    assert( row.yMax == 7 );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 7 );
    assert( row.xMax == 7 );
    assert( row.yMax == 7 );
}

// traverse a rectanular reange
void wzPatternIteratorTest::testVertical5box()
{
    wzVertical5 v5;
    wzPatternIterator& it = v5.GetIterator(); 

    // test SetBound(): a point
    wzBound box;
    box.xMin = 7;
    box.yMin = 7;
    box.xMax = 7;
    box.yMax = 7;
    assert( !it.SetBound( box ) );

    // test SetBound(): not tall enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 2;
    box.yMax = 5;
    assert( !it.SetBound( box ) );

    // test SetBound(): not tall enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 6;
    box.yMax = 4;
    assert( !it.SetBound( box ) );

    // test SetBound(): just minimal size
    box.xMin = 5;
    box.yMin = 5;
    box.xMax = 5;
    box.yMax = 9;
    assert( it.SetBound( box ) );

    //
    // teat a traverse range: (5,4) - (6,9)
    //
    box.xMin = 5;
    box.yMin = 4;
    box.xMax = 6;
    box.yMax = 9;
    assert( it.SetBound( box ) );

    // 1st element
    wzBound row = it.GetRow();
    assert( row.xMin == 5 );
    assert( row.yMin == 4 );
    assert( row.xMax == 5 );
    assert( row.yMax == 8 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 5 );
    assert( row.yMin == 5 );
    assert( row.xMax == 5 );
    assert( row.yMax == 9 );

    // 4th element
    assert( it.Next() );
    const wzBitBoard&  p = it.GetPattern();
    assert( !p.IsEmpty( 6, 5 ) );
    assert( !p.IsEmpty( 6, 6 ) );
    assert( !p.IsEmpty( 6, 7 ) );
    assert( !p.IsEmpty( 6, 8 ) );
    assert( !p.IsEmpty( 6, 9 ) );

    // only 5 bits
    assert( (p & p) == 5 );

    // others are empty, try some of them
    assert( p.IsEmpty( 5, 5 ) );
    assert( p.IsEmpty( 5, 6 ) );
    assert( p.IsEmpty( 5, 7 ) );
    assert( p.IsEmpty( 5, 8 ) );
    assert( p.IsEmpty( 5, 9 ) );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    assert( row.xMin == 5 );
    assert( row.yMin == 4 );
    assert( row.xMax == 5 );
    assert( row.yMax == 8 );
}

// traverse a line
void wzPatternIteratorTest::testVertical5line()
{
    wzVertical5 v5;
    wzPatternIterator& it = v5.GetIterator(); 

    //
    // one element
    //
    assert( it.SetCenter( 3, 0 ) );
    wzBound row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 0 );
    assert( row.xMax == 3 );
    assert( row.yMax == 4 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 10, 14 ) );
    row = it.GetRow();
    assert( row.xMin == 10 );
    assert( row.yMin == 10 );
    assert( row.xMax == 10 );
    assert( row.yMax == 14 );

    // no more
    assert( !it.Next() );

    //
    // 5 elements
    //
    assert( it.SetCenter( 7, 7 ) );

    // 1st element
    row = it.GetRow();
    assert( row.xMin == 7 );
    assert( row.yMin == 3 );
    assert( row.xMax == 7 );
    assert( row.yMax == 7 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );

    // 4th element
    assert( it.Next() );

    // 5th element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 7 );
    assert( row.yMin == 7 );
    assert( row.xMax == 7 );
    assert( row.yMax == 11 );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    assert( row.xMin == 7 );
    assert( row.yMin == 3 );
    assert( row.xMax == 7 );
    assert( row.yMax == 7 );
}

// traverse a rectanular reange
void wzPatternIteratorTest::testDiagonal5Fbox()
{
    wzDiagonal5F d5f;
    wzPatternIterator& it = d5f.GetIterator(); 

    // test SetBound(): a point
    wzBound box;
    box.xMin = 7;
    box.yMin = 7;
    box.xMax = 7;
    box.yMax = 7;
    assert( !it.SetBound( box ) );

    // test SetBound(): not wild enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 5;
    box.yMax = 2;
    assert( !it.SetBound( box ) );

    // test SetBound(): not wild enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 3;
    box.yMax = 8;
    assert( !it.SetBound( box ) );

    // test SetBound(): just minimal size
    box.xMin = 5;
    box.yMin = 5;
    box.xMax = 9;
    box.yMax = 9;
    assert( it.SetBound( box ) );

    //
    // teat a traverse range: (4,3) - (9,8)
    //
    box.xMin = 4;
    box.yMin = 3;
    box.xMax = 9;
    box.yMax = 8;
    assert( it.SetBound( box ) );

    // 1st element
    wzBound row = it.GetRow();
    assert( row.xMin == 4 );
    assert( row.yMin == 3 );
    assert( row.xMax == 8 );
    assert( row.yMax == 7 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 4 );
    assert( row.yMin == 4 );
    assert( row.xMax == 8 );
    assert( row.yMax == 8 );

    // 4th element
    assert( it.Next() );
    const wzBitBoard&  p = it.GetPattern();
    assert( !p.IsEmpty( 5, 4 ) );
    assert( !p.IsEmpty( 6, 5 ) );
    assert( !p.IsEmpty( 7, 6 ) );
    assert( !p.IsEmpty( 8, 7 ) );
    assert( !p.IsEmpty( 9, 8 ) );

    // only 5 bits
    assert( (p & p) == 5 );

    // others are empty, try some of them
    assert( p.IsEmpty( 4, 4 ) );
    assert( p.IsEmpty( 5, 5 ) );
    assert( p.IsEmpty( 6, 6 ) );
    assert( p.IsEmpty( 7, 7 ) );
    assert( p.IsEmpty( 8, 8 ) );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    assert( row.xMin == 4 );
    assert( row.yMin == 3 );
    assert( row.xMax == 8 );
    assert( row.yMax == 7 );
}

// traverse a line
void wzPatternIteratorTest::testDiagonal5Fline()
{
    wzDiagonal5F d5f;
    wzPatternIterator& it = d5f.GetIterator(); 

    //
    // top-right corner: zero element
    //
    assert( !it.SetCenter( 11, 0 ) );
    assert( !it.SetCenter( 12, 0 ) );
    assert( !it.SetCenter( 13, 0 ) );
    assert( !it.SetCenter( 14, 0 ) );
    assert( !it.SetCenter( 12, 1 ) );
    assert( !it.SetCenter( 13, 1 ) );
    assert( !it.SetCenter( 14, 1 ) );
    assert( !it.SetCenter( 13, 2 ) );
    assert( !it.SetCenter( 14, 2 ) );
    assert( !it.SetCenter( 14, 3 ) );

    //
    // bottom-left corner: zero element
    //
    assert( !it.SetCenter( 0, 14 ) );
    assert( !it.SetCenter( 1, 14 ) );
    assert( !it.SetCenter( 2, 14 ) );
    assert( !it.SetCenter( 3, 14 ) );
    assert( !it.SetCenter( 0, 13 ) );
    assert( !it.SetCenter( 1, 13 ) );
    assert( !it.SetCenter( 2, 13 ) );
    assert( !it.SetCenter( 0, 12 ) );
    assert( !it.SetCenter( 1, 12 ) );
    assert( !it.SetCenter( 0, 11 ) );

    //
    // one element
    //
    assert( it.SetCenter( 3, 0 ) );
    wzBound row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 0 );
    assert( row.xMax == 7 );
    assert( row.yMax == 4 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 14, 6 ) );
    row = it.GetRow();
    assert( row.xMin == 10 );
    assert( row.yMin == 2 );
    assert( row.xMax == 14 );
    assert( row.yMax == 6 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 10, 14 ) );
    row = it.GetRow();
    assert( row.xMin == 6 );
    assert( row.yMin == 10 );
    assert( row.xMax == 10 );
    assert( row.yMax == 14 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 0, 5 ) );
    row = it.GetRow();
    assert( row.xMin == 0 );
    assert( row.yMin == 5 );
    assert( row.xMax == 4 );
    assert( row.yMax == 9 );

    // no more
    assert( !it.Next() );

    //
    // 2 elements
    //
    assert( it.SetCenter( 3, 1 ) );
    row = it.GetRow();
    assert( row.xMin == 2 );
    assert( row.yMin == 0 );
    assert( row.xMax == 6 );
    assert( row.yMax == 4 );

    // 2nd element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 1 );
    assert( row.xMax == 7 );
    assert( row.yMax == 5 );

    // no more
    assert( !it.Next() );

    //
    // 5 elements
    //
    assert( it.SetCenter( 7, 7 ) );

    // 1st element
    row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 3 );
    assert( row.xMax == 7 );
    assert( row.yMax == 7 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );

    // 4th element
    assert( it.Next() );

    // 5th element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 7 );
    assert( row.yMin == 7 );
    assert( row.xMax == 11 );
    assert( row.yMax == 11 );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 3 );
    assert( row.xMax == 7 );
    assert( row.yMax == 7 );
}

// traverse a rectanular reange
void wzPatternIteratorTest::testDiagonal5Bbox()
{
    wzDiagonal5B d5b;
    wzPatternIterator& it = d5b.GetIterator(); 

    // test SetBound(): a point
    wzBound box;
    box.xMin = 7;
    box.yMin = 7;
    box.xMax = 7;
    box.yMax = 7;
    assert( !it.SetBound( box ) );

    // test SetBound(): not wild enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 5;
    box.yMax = 2;
    assert( !it.SetBound( box ) );

    // test SetBound(): not wild enough
    box.xMin = 2;
    box.yMin = 2;
    box.xMax = 3;
    box.yMax = 8;
    assert( !it.SetBound( box ) );

    // test SetBound(): just minimal size
    box.xMin = 5;
    box.yMin = 5;
    box.xMax = 9;
    box.yMax = 9;
    assert( it.SetBound( box ) );

    //
    // teat a traverse range: (4,3) - (9,8)
    //
    box.xMin = 4;
    box.yMin = 3;
    box.xMax = 9;
    box.yMax = 8;
    assert( it.SetBound( box ) );

    // 1st element
    wzBound row = it.GetRow();
    // not min and max, but start and end point
    assert( row.xMin == 4 );
    assert( row.yMin == 7 );
    assert( row.xMax == 8 );
    assert( row.yMax == 3 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );
    row = it.GetRow();
    // not min and max, but start and end point
    assert( row.xMin == 4 );
    assert( row.yMin == 8 );
    assert( row.xMax == 8 );
    assert( row.yMax == 4 );

    // 4th element
    assert( it.Next() );
    const wzBitBoard&  p = it.GetPattern();
    assert( !p.IsEmpty( 5, 8 ) );
    assert( !p.IsEmpty( 6, 7 ) );
    assert( !p.IsEmpty( 7, 6 ) );
    assert( !p.IsEmpty( 8, 5 ) );
    assert( !p.IsEmpty( 9, 4 ) );

    // only 5 bits
    assert( (p & p) == 5 );

    // others are empty, try some of them
    assert( p.IsEmpty( 4, 4 ) );
    assert( p.IsEmpty( 5, 5 ) );
    assert( p.IsEmpty( 6, 6 ) );
    assert( p.IsEmpty( 7, 7 ) );
    assert( p.IsEmpty( 8, 8 ) );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    // not min and max, but start and end point
    assert( row.xMin == 4 );
    assert( row.yMin == 7 );
    assert( row.xMax == 8 );
    assert( row.yMax == 3 );
}

// traverse a line
void wzPatternIteratorTest::testDiagonal5Bline()
{
    wzDiagonal5B d5b;
    wzPatternIterator& it = d5b.GetIterator(); 

    //
    // top-left corner: zero element
    //
    assert( !it.SetCenter( 0, 0 ) );
    assert( !it.SetCenter( 1, 0 ) );
    assert( !it.SetCenter( 2, 0 ) );
    assert( !it.SetCenter( 3, 0 ) );
    assert( !it.SetCenter( 0, 1 ) );
    assert( !it.SetCenter( 1, 1 ) );
    assert( !it.SetCenter( 2, 1 ) );
    assert( !it.SetCenter( 0, 2 ) );
    assert( !it.SetCenter( 1, 2 ) );
    assert( !it.SetCenter( 0, 3 ) );

    //
    // bottom-right corner: zero element
    //
    assert( !it.SetCenter( 11, 14 ) );
    assert( !it.SetCenter( 12, 14 ) );
    assert( !it.SetCenter( 13, 14 ) );
    assert( !it.SetCenter( 14, 14 ) );
    assert( !it.SetCenter( 12, 13 ) );
    assert( !it.SetCenter( 13, 13 ) );
    assert( !it.SetCenter( 14, 13 ) );
    assert( !it.SetCenter( 13, 12 ) );
    assert( !it.SetCenter( 14, 12 ) );
    assert( !it.SetCenter( 14, 11 ) );

    //
    // one element
    //
    assert( it.SetCenter( 5, 0 ) );
    wzBound row = it.GetRow();
    assert( row.xMin == 1 );
    assert( row.yMin == 4 );
    assert( row.xMax == 5 );
    assert( row.yMax == 0 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 14, 6 ) );
    row = it.GetRow();
    assert( row.xMin == 10 );
    assert( row.yMin == 10 );
    assert( row.xMax == 14 );
    assert( row.yMax == 6 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 8, 14 ) );
    row = it.GetRow();
    assert( row.xMin == 8 );
    assert( row.yMin == 14 );
    assert( row.xMax == 12 );
    assert( row.yMax == 10 );

    // no more
    assert( !it.Next() );

    //
    // one element
    //
    assert( it.SetCenter( 0, 5 ) );
    row = it.GetRow();
    assert( row.xMin == 0 );
    assert( row.yMin == 5 );
    assert( row.xMax == 4 );
    assert( row.yMax == 1 );

    // no more
    assert( !it.Next() );

    //
    // 2 elements
    //
    assert( it.SetCenter( 4, 1 ) );
    row = it.GetRow();
    assert( row.xMin == 1 );
    assert( row.yMin == 4 );
    assert( row.xMax == 5 );
    assert( row.yMax == 0 );

    // 2nd element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 0 );
    assert( row.yMin == 5 );
    assert( row.xMax == 4 );
    assert( row.yMax == 1 );

    // no more
    assert( !it.Next() );

    //
    // 5 elements
    //
    assert( it.SetCenter( 7, 7 ) );
    row = it.GetRow();
    assert( row.xMin == 7 );
    assert( row.yMin == 7 );
    assert( row.xMax == 11 );
    assert( row.yMax == 3 );

    // 2nd element
    assert( it.Next() );

    // 3rd element
    assert( it.Next() );

    // 4th element
    assert( it.Next() );

    // 5th element
    assert( it.Next() );
    row = it.GetRow();
    assert( row.xMin == 3 );
    assert( row.yMin == 11 );
    assert( row.xMax == 7 );
    assert( row.yMax == 7 );

    // no more
    assert( !it.Next() );

    // reset and point to 1st element again.
    it.Reset();
    row = it.GetRow();
    assert( row.xMin == 7 );
    assert( row.yMin == 7 );
    assert( row.xMax == 11 );
    assert( row.yMax == 3 );
}

Test* wzPatternIteratorTest::suite()
{
    TestSuite *suite = new TestSuite( "wzPatternIteratorTest" );

    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testHorizontal5box ) );
    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testHorizontal5line ) );
    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testVertical5box ) );
    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testVertical5line ) );
    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testDiagonal5Fbox ) );
    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testDiagonal5Fline ) );
    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testDiagonal5Bbox ) );
    suite->addTest( CU_SUIT_TEST( wzPatternIteratorTest, testDiagonal5Bline ) );

    return suite;
}
